# Generated by Django 4.1.8 on 2023-05-19 11:23

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0047_alter_episode_podcast_audio"),
    ]

    operations = [
        migrations.AlterField(
            model_name="post",
            name="visible_date",
            field=models.DateTimeField(
                db_index=True,
                default=django.utils.timezone.now,
                help_text="The visible date of the post which is used for sorting.",
            ),
        ),
    ]



#!/usr/bin/env python

import os
import shutil
import sys

from pathlib import Path

import django

from django.conf import settings
from django.test.utils import get_runner


def run_tests(*test_args):
    if not test_args:
        test_args = ["tests"]

    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(test_args)
    media_root = Path(settings.MEDIA_ROOT)
    try:
        shutil.rmtree(media_root)  # FIXME move cleanup to tests
    except FileNotFoundError:
        pass
    sys.exit(bool(failures))


if __name__ == "__main__":
    run_tests(*sys.argv[1:])



from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render

from ..forms import SelectThemeForm
from ..models import get_template_base_dir
from .htmx_helpers import HtmxHttpRequest


def set_template_base_dir(request: HtmxHttpRequest, template_base_dir: str) -> None:
    """Store the template base dir in the session."""
    request.session["template_base_dir"] = template_base_dir


def select_theme(request: HtmxHttpRequest) -> HttpResponse:
    """
    Store the selected theme in the session. This is used to
    determine the template base directory for rendering the
    blog and episode pages.
    """
    template_base_dir = get_template_base_dir(request, None)
    # use the referer as the next url if it exists because request.path is
    # the url of the select theme view, and we want to redirect to the previous page
    next_url = request.headers.get("referer", request.path)
    if request.method == "POST":
        form = SelectThemeForm(request.POST)
        if form.is_valid():
            set_template_base_dir(request, form.cleaned_data["template_base_dir"])
            success_url = form.cleaned_data["next"]
            return HttpResponseRedirect(success_url)
    else:
        form = SelectThemeForm(
            initial={
                "template_base_dir": template_base_dir,
                "next": next_url,
            }
        )
    context = {
        "theme_form": form,
        "template_base_dir": template_base_dir,
        "template_base_dir_choices": form.fields["template_base_dir"].choices,  # type: ignore
        "next_url": next_url,
    }
    return render(request, f"cast/{template_base_dir}/select_theme.html", context=context)



import os
import platform
import subprocess
import sys
import webbrowser
from pathlib import Path


def get_project_root():
    return Path(__file__).parent.resolve()


def bootstrap():
    """
    Called when first non-standard lib import fails.

    We need to install the development requirements to use this script.
    """

    def get_base_prefix_compat():
        """Get base/real prefix, or sys.prefix if there is none."""
        return getattr(sys, "base_prefix", None) or getattr(sys, "real_prefix", None) or sys.prefix

    def in_virtualenv():
        return get_base_prefix_compat() != sys.prefix

    def pip_install(package):
        subprocess.run([sys.executable, "-m", "pip", "install", package], check=True)

    if not in_virtualenv():
        print("Please create a virtual environment first and activate it!")
        sys.exit(1)

    pip_install("flit")

    print("Empty virtualenv, installing development dependencies..")
    subprocess.run([sys.executable, "-m", "flit", "install", "-s"], check=True)


try:
    import typer
except ImportError:
    bootstrap()
    import typer

from rich import print  # noqa

cli = typer.Typer()


def get_pythonpath():
    """Add project root and model directory to string"""
    project_root = str(get_project_root())
    model_root = str(Path(__file__).parent / "model")
    return f"{project_root}:{model_root}"


def env_with_pythonpath():
    """Get en environment dict with includes PYTHONPATH"""
    env = os.environ.copy()
    env["PYTHONPATH"] = get_pythonpath()
    return env


@cli.command()
def mypy():
    """Run Mypy (configured in pyproject.toml)"""
    subprocess.run(["mypy", "cast"])


@cli.command()
def test(test_path: str = typer.Argument(None)):
    if test_path is None:
        test_path = "tests"
    subprocess.call([sys.executable, "runtests.py", test_path])
    # FIXME use pytest after fixing removing test images after running tests
    # subprocess.call(["python", "-m", "pytest"], env=env_with_pythonpath())


@cli.command()
def coverage():
    """
    Run and show coverage.
    """
    subprocess.call(["coverage", "run", "-m", "pytest"], env=env_with_pythonpath())
    subprocess.call(["coverage", "html"])
    if platform.system() == "Darwin":
        subprocess.call(["open", "htmlcov/index.html"])
    elif platform.system() == "Linux" and "Microsoft" in platform.release():  # on WSL
        subprocess.call(["explorer.exe", r"htmlcov\index.html"])


@cli.command()
def jupyterlab():
    """
    Start a jupyterlab server.
    """
    project_root = get_project_root()
    notebook_dir = project_root / "notebooks"
    notebook_dir.mkdir(exist_ok=True)
    env = env_with_pythonpath() | {"DJANGO_ALLOW_ASYNC_UNSAFE": "true"}
    subprocess.call([sys.executable, "example/manage.py", "shell_plus", "--notebook"], env=env)


@cli.command()
def docs():
    autogenerated = [
        "cast.api.rst",
        "cast.migrations.rst",
        "cast.rst",
        "cast.templatetags.rst",
        "modules.rst",
    ]
    for rst_name in autogenerated:
        (Path("docs") / rst_name).unlink(missing_ok=True)
    commands = [
        # ["sphinx-apidoc", "-o", "docs/", "cast"],
        ["make", "-C", "docs", "clean"],
        ["make", "-C", "docs", "html"],
    ]
    for command in commands:
        subprocess.call(command)
    file_url = "file://" + str(Path("docs/_build/html/index.html").resolve())
    webbrowser.open_new_tab(file_url)


if __name__ == "__main__":
    cli()



from django.apps import apps


def test_delete_wagtail_images_app_setting(mocker):
    cast_config = apps.get_app_config("cast")

    # delete images True -> do not disconnect post_delete_file_cleanup
    mocker.patch("cast.appsettings.DELETE_WAGTAIL_IMAGES", True)
    mocked_disconnect = mocker.patch("cast.appsettings.post_delete.disconnect")
    cast_config.ready()
    assert mocked_disconnect.call_count == 0

    # delete images False -> disconnect post_delete_file_cleanup
    mocker.patch("cast.appsettings.DELETE_WAGTAIL_IMAGES", False)
    mocked_disconnect = mocker.patch("cast.appsettings.post_delete.disconnect")
    cast_config.ready()
    assert mocked_disconnect.call_count == 1



from .audio import Audio, ChapterMark, sync_chapter_marks
from .file import File
from .gallery import Gallery, get_or_create_gallery
from .index_pages import Blog, Podcast
from .itunes import ItunesArtWork
from .moderation import SpamFilter
from .pages import Episode, HomePage, Post, sync_media_ids
from .snippets import PostCategory
from .theme import (
    TemplateBaseDirectory,
    get_template_base_dir,
    get_template_base_dir_choices,
)
from .video import Video, get_video_dimensions

__all__ = [
    "Audio",
    "ChapterMark",
    "sync_chapter_marks",
    "Blog",
    "File",
    "Gallery",
    "get_or_create_gallery",
    "get_template_base_dir",
    "get_template_base_dir_choices",
    "HomePage",
    "ItunesArtWork",
    "Post",
    "PostCategory",
    "Episode",
    "sync_media_ids",
    "SpamFilter",
    "Video",
    "get_video_dimensions",
    "Podcast",
    "TemplateBaseDirectory",
]



from pathlib import Path
from typing import Union

from django.conf import settings
from django.db import models
from django.template import engines
from django.template.loaders.base import Loader as BaseLoader
from django.utils.translation import gettext_lazy as _
from wagtail.contrib.settings.models import BaseSiteSetting, register_setting

from ..views import HtmxHttpRequest


def get_required_template_names() -> list[str]:
    """
    Return a list of template names that have to be present in a theme. Not just
    a constant because maybe it's possible to calculate this dynamically in the
    future.
    """
    return [
        "blog_list_of_posts.html",
        "post.html",
        "post_body.html",
        "episode.html",
    ]


def is_compatible_template_loader(template_loader: BaseLoader) -> bool:
    """
    Return True if the template loader is compatible with theme discovery.
    """
    return hasattr(template_loader, "get_dirs")


def get_template_directories() -> list[Path]:
    """
    Return a list of template directories from all template loaders.
    Only filesystem loaders are supported.
    """
    template_directories = []
    for engine in engines.all():
        for template_loader in engine.engine.template_loaders:  # type: ignore
            if is_compatible_template_loader(template_loader):
                for template_directory in template_loader.get_dirs():
                    if isinstance(template_directory, str):
                        template_directory = Path(template_directory)
                    template_directories.append(template_directory)
    return template_directories


def is_cast_template_base_dir_candidate(base_dir: Path, required_template_names: list[str]) -> bool:
    """
    Return True if base_dir is a candidate for a template base directory which means
    it contains all required template files from required_template_names.
    """
    for template_name in required_template_names:
        if not (base_dir / template_name).exists():
            # a required template does not exist -> base_dir is not a candidate
            return False
    return True


def get_template_base_dir_candidates(
    template_directories: list[Path], required_template_names: list[str] = get_required_template_names()
) -> list[str]:
    """
    Return a list of template base directory candidates containing the required
    template file names.
    """
    directory_matches: set[Path] = set()
    if len(required_template_names) == 0:
        # nothing required -> nothing to find -> return early
        return []
    required_template_name = required_template_names[0]
    for template_directory in template_directories:
        for directory_match in list(template_directory.glob(f"**/cast/**/{required_template_name}")):
            directory_matches.add(directory_match.parent)
    return [dm.name for dm in directory_matches if is_cast_template_base_dir_candidate(dm, required_template_names)]


class TemplateName(models.TextChoices):
    BOOTSTRAP4 = "bootstrap4", _("Bootstrap 4")
    PLAIN = "plain", _("Just HTML")


def get_template_base_dir_choices() -> list[tuple[str, str]]:
    """
    Return a list of choices for the template base directory setting.
    """
    # handle predefined choices
    choices, seen = [], set()
    for template_name in TemplateName:
        choices.append((template_name.value, template_name.label))
        seen.add(template_name.value)

    # handle custom choices via settings
    for template_name, display_name in getattr(settings, "CAST_CUSTOM_THEMES", []):
        if template_name not in seen:
            choices.append((template_name, display_name))
            seen.add(template_name)

    # search for template base directories
    template_directories = get_template_directories()
    template_base_dir_candidates = get_template_base_dir_candidates(template_directories)
    for candidate in template_base_dir_candidates:
        if candidate not in seen:
            choices.append((candidate, candidate))

    return choices


def get_template_base_dir(request: HtmxHttpRequest, pre_selected: Union[str, None]) -> str:
    if hasattr(request, "session") and (template_base_dir := request.session.get("template_base_dir")) is not None:
        return template_base_dir
    if pre_selected is not None:
        return pre_selected
    else:
        return TemplateBaseDirectory.for_request(request).name


@register_setting
class TemplateBaseDirectory(BaseSiteSetting):
    """
    The base directory for templates. Makes it possible to use different
    templates for different sites / change look and feel of the site from
    the wagtail admin.
    """

    name = models.CharField(
        choices=get_template_base_dir_choices(),
        max_length=128,
        default=TemplateName.BOOTSTRAP4,
        help_text=_(
            "The theme to use for this site implemented as a template base directory. "
            "It's possible to overwrite this setting for each blog."
            "If you want to use a custom theme, you have to create a new directory "
            "in your template directory named cast/<your-theme-name>/ and put all "
            "required templates in there."
        ),
    )



# Generated by Django 4.1.4 on 2023-02-03 06:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0031_remove_timestampedmodel_because_wagtail"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="blog",
            name="created",
        ),
        migrations.RemoveField(
            model_name="blog",
            name="modified",
        ),
    ]



from datetime import timedelta

import pytest
from django.urls import reverse
from django.utils import timezone

from cast.api.views import (
    AudioPodloveDetailView,
    FilteredPagesAPIViewSet,
    ThemeListView,
)
from cast.models import PostCategory

from .factories import UserFactory


def test_api_root(api_client):
    """Test that the API root returns a 200."""
    url = reverse("cast:api:root")
    r = api_client.get(url)
    assert r.status_code == 200


class TestBlogVideo:
    pytestmark = pytest.mark.django_db

    @classmethod
    def setup_class(cls):
        cls.list_url = reverse("cast:api:video_list")
        cls.detail_url = reverse("cast:api:video_detail", kwargs={"pk": 1})

    def test_video_list_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the list
        endpoint without being authenticated.
        """
        r = api_client.get(self.list_url, format="json")
        assert r.status_code == 403

    def test_video_detail_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the
        detail endpoint without being authenticated.
        """
        r = api_client.get(self.detail_url, format="json")
        assert r.status_code == 403

    def test_video_list_endpoint_with_authentication(self, api_client):
        """Check for list result when accessing the list endpoint
        being logged in.
        """
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.list_url, format="json")
        # dont redirect to login page
        assert r.status_code == 200
        assert "results" in r.json()


class TestBlogAudio:
    pytestmark = pytest.mark.django_db

    @classmethod
    def setup_class(cls):
        cls.list_url = reverse("cast:api:audio_list")
        cls.detail_url = reverse("cast:api:audio_detail", kwargs={"pk": 1})

    def test_audio_list_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the list
        endpoint without being authenticated.
        """
        r = api_client.get(self.list_url, format="json")
        assert r.status_code == 403

    def test_audio_detail_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the
        detail endpoint without being authenticated.
        """
        r = api_client.get(self.detail_url, format="json")
        assert r.status_code == 403

    def test_audio_list_endpoint_with_authentication(self, api_client):
        """Check for list result when accessing the list endpoint
        being logged in.
        """
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.list_url, format="json")
        # dont redirect to login page
        assert r.status_code == 200
        assert "results" in r.json()


class TestPodcastAudio:
    pytestmark = pytest.mark.django_db

    def test_podlove_detail_endpoint_without_authentication(self, api_client, audio):
        """Should be accessible without authentication."""
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        assert r.status_code == 200

    def test_podlove_detail_endpoint_duration(self, api_client, audio):
        """Test whether microseconds get stripped away from duration via api - they have
        to be for podlove player to work.
        """
        delta = timedelta(days=0, hours=1, minutes=10, seconds=20, microseconds=40)
        audio.duration = delta
        audio.save()
        assert "." in str(audio.duration)
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        assert "." not in r.json()["duration"]

    def test_podlove_detail_endpoint_includes_link_to_episode(self, api_client, episode):
        """Test whether the podlove detail endpoint includes a link to the episode."""
        audio = episode.podcast_audio
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})

        r = api_client.get(podlove_detail_url, format="json")
        assert r.status_code == 200

        # link is always included, might be empty
        assert "link" in r.json()

        # explicitly set episode_id FIXME: only works if there are multiple episodes for audio
        podlove_detail_url_with_episode_id = f"{podlove_detail_url}?episode_id={episode.pk}"

        r = api_client.get(podlove_detail_url_with_episode_id, format="json")
        assert r.status_code == 200

        podlove_data = r.json()
        assert "link" in podlove_data
        assert podlove_data["link"] == episode.full_url

    def test_podlove_detail_endpoint_chaptermarks(self, api_client, audio, chaptermarks):
        """Test whether chaptermarks get delivered via podlove endpoint."""
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        chapters = r.json()["chapters"]
        for chapter in chapters:
            # assert microseconds are stripped away
            assert "." not in chapter["start"]
            assert ":" in chapter["start"]
        assert len(chapters) == 3
        # assert reordering
        assert chapters[-1]["title"] == "coughing"

    def test_podlove_detail_retrieve_with_value_error(self, mocker):
        class MockRequest:
            query_params = {"episode_id": "foo"}

        mocker.patch("cast.api.views.AudioPodloveDetailView.get_object")
        mocker.patch("cast.api.views.AudioPodloveDetailView.get_serializer")
        podlove_view = AudioPodloveDetailView()
        response = podlove_view.retrieve(MockRequest())
        assert response.status_code == 200


class TestCommentTrainingData:
    pytestmark = pytest.mark.django_db

    @classmethod
    def setup_class(cls):
        cls.url = reverse("cast:api:comment-training-data")

    def test_get_comment_training_data_without_authentication(self, api_client):
        """Should not be accessible without authentication."""
        r = api_client.get(self.url, format="json")
        assert r.status_code == 403

    def test_get_comment_training_data_with_authentication(self, api_client):
        """Check for list result when accessing the training data endpoint being logged in."""
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.url, format="json")
        assert r.status_code == 200
        assert r.json() == []


@pytest.mark.django_db
@pytest.mark.parametrize(
    "date, post_filter, len_result",
    [
        (timezone.datetime(2022, 8, 22), "true", 0),  # wrong date facet -> not found
        (timezone.now(), "true", 1),  # correct date facet -> found
        (timezone.datetime(2022, 8, 22), "false", 1),  # wrong date facet and no post filter -> found
    ],
)
def test_wagtail_pages_api_with_post_filter(date, post_filter, len_result, rf, blog, post):
    viewset = FilteredPagesAPIViewSet()
    path = blog.wagtail_api_pages_url
    date_facet = f"{date.year}-{date.month}"
    request = rf.get(
        f"{path}?child_of={blog.pk}&type=cast.Post&date_facets={date_facet}&use_post_filter={post_filter}"
    )
    viewset.request = request
    queryset = viewset.get_queryset()
    assert len(queryset) == len_result


@pytest.mark.django_db
def test_wagtail_pages_api_with_post_filter_and_fulltext_search(rf, blog, post):
    viewset = FilteredPagesAPIViewSet()
    path = blog.wagtail_api_pages_url
    search_param = "search=foo"
    request = rf.get(f"{path}?child_of={blog.pk}&type=cast.Post&{search_param}&use_post_filter=true")
    viewset.request = request
    queryset = viewset.get_queryset()
    assert len(queryset) == 0


@pytest.mark.django_db
def test_facet_counts_list(api_client, blog):
    """
    Test whether the facet counts list endpoint returns a list of all blogs
    and the blog is included in the result.
    """
    url = reverse("cast:api:facet-counts-list")
    r = api_client.get(url, format="json")
    assert r.status_code == 200

    [result] = r.json()["results"]
    assert "id" in result
    assert "url" in result
    assert result["id"] == blog.pk


@pytest.mark.django_db
def test_facet_counts_detail(api_client, blog, post):
    """
    Test whether the facet counts detail endpoint returns the
    facet counts for a specific blog.
    """
    # Given a post with a category and a tag
    category = PostCategory.objects.create(name="category", slug="category")
    post.categories.add(category)
    post.tags.add("tag")
    post.save()

    # When we request the facet counts for the blog
    url = reverse("cast:api:facet-counts-detail", kwargs={"pk": blog.pk})
    r = api_client.get(url, format="json")
    assert r.status_code == 200

    result = r.json()
    facet_counts = result["facet_counts"]

    # Then we expect the correct facet counts to be returned
    assert facet_counts["date_facets"][0]["slug"] == post.visible_date.strftime("%Y-%m")
    assert facet_counts["date_facets"][0]["name"] == post.visible_date.strftime("%Y-%m")
    assert facet_counts["date_facets"][0]["count"] == 1

    assert facet_counts["category_facets"][0]["slug"] == category.slug
    assert facet_counts["category_facets"][0]["count"] == 1

    assert facet_counts["tag_facets"][0]["slug"] == "tag"
    assert facet_counts["tag_facets"][0]["count"] == 1

    # make sure adding a search param filters the results
    r = api_client.get(f"{url}?search=foobar", format="json")
    assert r.status_code == 200

    result = r.json()
    date_facets = result["facet_counts"]["date_facets"]
    assert len(date_facets) == 0


@pytest.mark.django_db
def test_get_comments_via_post_detail(api_client, post, comment):
    url = reverse("cast:api:wagtail:pages:detail", kwargs={"pk": post.pk})
    r = api_client.get(url, format="json")
    assert r.status_code == 200

    comments = r.json()["comments"]
    assert comments[0]["comment"] == comment.comment


def test_theme_list_queryset_is_none():
    view = ThemeListView()
    assert view.get_queryset() is None


@pytest.mark.django_db
def test_list_themes(api_client):
    # Given an api url to fetch the list of themes
    url = reverse("cast:api:theme-list")
    # When we request the list of themes
    r = api_client.get(url, format="json")
    assert r.status_code == 200

    # Then we expect a list of themes to be returned and include the `plain` theme
    result = r.json()
    assert "plain" in {theme["slug"] for theme in result["items"]}


@pytest.mark.django_db
def test_update_theme(api_client):
    # Given an api url to update the theme
    url = reverse("cast:api:theme-update")
    # When we post to the update theme endpoint
    r = api_client.post(url, {"theme_slug": "plain"}, format="json")
    assert r.status_code == 200

    # Then we expect a success message to be returned
    result = r.json()
    assert result["message"] == "Theme updated successfully"
    assert api_client.session.get("template_base_dir") == "plain"


@pytest.mark.django_db
def test_update_theme_invalid(api_client):
    # Given an api url to update the theme
    url = reverse("cast:api:theme-update")
    # When we post an invalid theme to the update theme endpoint
    r = api_client.post(url, {"theme_slug": "invalid"}, format="json")
    assert r.status_code == 400

    # Then we expect an error message to be returned and
    # the theme is not stored in the session
    result = r.json()
    print(result)
    assert result["error"] == "Theme slug is invalid"
    assert api_client.session.get("template_base_dir") is None


@pytest.mark.django_db
def test_render_html_with_theme_from_session(api_client, post):
    # Given we have custom theme set in the session
    r = api_client.post(
        # FIXME there's some way to update the session more elegantly
        # use this instead of the post request
        reverse("cast:api:theme-update"),
        {"theme_slug": "plain"},
        format="json",
    )
    assert r.status_code == 200

    # When we request the blog post via api
    url = reverse("cast:api:wagtail:pages:detail", kwargs={"pk": post.pk})
    r = api_client.get(url, format="json")
    assert r.status_code == 200

    # Then we expect the blog post to be rendered with the theme from the session
    assert r.context.get("template_base_dir") == "plain"
    assert all([t.name.startswith("cast/plain/") for t in r.templates])



# Generated by Django 3.2.6 on 2021-08-30 09:22

import django.db.models.deletion
import taggit.managers
import wagtail.models.collections
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taggit", "0003_taggeditem_add_unique_index"),
        ("wagtailcore", "0062_comment_models_and_pagesubscription"),
        ("cast", "0015_delete_blogindexpage"),
    ]

    operations = [
        migrations.AddField(
            model_name="audio",
            name="collection",
            field=models.ForeignKey(
                default=wagtail.models.collections.get_root_collection_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="wagtailcore.collection",
                verbose_name="collection",
            ),
        ),
        migrations.AddField(
            model_name="audio",
            name="tags",
            field=taggit.managers.TaggableManager(
                blank=True, help_text=None, through="taggit.TaggedItem", to="taggit.Tag", verbose_name="tags"
            ),
        ),
    ]



import json
import logging
import uuid
from datetime import datetime
from typing import Any

import django.forms.forms
from django.core.paginator import InvalidPage, Paginator
from django.db import models
from django.http import Http404
from django.http.request import QueryDict
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from wagtail.admin.panels import FieldPanel
from wagtail.api import APIField
from wagtail.fields import RichTextField
from wagtail.models import Page, PageManager

from cast import appsettings
from cast.filters import PostFilterset
from cast.models.itunes import ItunesArtWork

from ..views import HtmxHttpRequest
from .pages import Post
from .theme import get_template_base_dir, get_template_base_dir_choices

logger = logging.getLogger(__name__)


ContextDict = dict[str, Any]


class Blog(Page):
    """
    This is the index page for a blog. It contains a list of posts.
    """

    author = models.CharField(
        max_length=255,
        default=None,
        null=True,
        blank=True,
        help_text=_("Freeform text that will be used in the feed."),
    )
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    email = models.EmailField(null=True, default=None, blank=True)
    comments_enabled = models.BooleanField(
        _("comments_enabled"),
        default=True,
        help_text=_("Whether comments are enabled for this blog." ""),
    )
    noindex = models.BooleanField(
        "noindex",
        default=False,
        help_text=_(
            "Whether to add a noindex meta tag to this page and all subpages excluding them from search engines."
        ),
    )
    template_base_dir = models.CharField(
        choices=get_template_base_dir_choices(),
        max_length=128,
        default=None,
        null=True,
        blank=True,
        help_text=_(
            "The theme to use for this blog implemented as a template base directory. "
            "If not set, the template base directory will be determined by a site setting."
        ),
    )

    # wagtail
    description = RichTextField(blank=True)
    content_panels = Page.content_panels + [
        FieldPanel("description", classname="full"),
        FieldPanel("email"),
        FieldPanel("author"),
        FieldPanel("template_base_dir"),
    ]
    promote_panels = Page.promote_panels + [
        FieldPanel("noindex"),
    ]
    api_fields = [
        APIField("description"),
    ]

    subpage_types = ["cast.Post"]
    is_podcast = False

    def __str__(self):
        return self.title

    def get_template_base_dir(self, request: HtmxHttpRequest) -> str:
        return get_template_base_dir(request, self.template_base_dir)

    def get_template(self, request: HtmxHttpRequest, *args, **kwargs) -> str:
        template_base_dir = self.get_template_base_dir(request)
        template_name = "blog_list_of_posts.html"  # full template
        if request.htmx:
            target_to_template_name = {
                "paging-area": "_list_of_posts_and_paging_controls.html",
            }
            if request.htmx.target is not None:
                template_name = target_to_template_name[request.htmx.target]
            else:
                logger.warning("HTMX target is None")
        template = f"cast/{template_base_dir}/{template_name}"
        return template

    @property
    def last_build_date(self) -> datetime:
        return Post.objects.live().descendant_of(self.blog).order_by("-visible_date")[0].visible_date

    @property
    def author_name(self) -> str:
        if self.author is not None:
            return self.author
        else:
            return self.owner.get_full_name()

    @property
    def unfiltered_published_posts(self) -> models.QuerySet[Post]:
        if self.pk is None:
            # this blog is not saved to database yet, therefore it has no posts
            return Post.objects.none()
        return Post.objects.live().descendant_of(self).order_by("-visible_date")

    def get_filterset(self, get_params: QueryDict) -> PostFilterset:
        return PostFilterset(data=get_params, queryset=self.unfiltered_published_posts)

    @staticmethod
    def get_published_posts(filtered_posts: models.QuerySet) -> models.QuerySet[Post]:
        queryset = filtered_posts
        queryset = queryset.select_related("owner")
        queryset = queryset.prefetch_related(
            "audios",
            "images",
            "videos",
            "galleries",
            "galleries__images",
            "images__renditions",
            "galleries__images__renditions",
        )
        return queryset

    @staticmethod
    def paginate_queryset(context: ContextDict, posts_queryset: models.QuerySet, get_params: QueryDict) -> ContextDict:
        paginator = Paginator(posts_queryset, appsettings.POST_LIST_PAGINATION)
        page_from_url = "1"
        if "page" in get_params:
            page_from_url = str(get_params["page"])
        try:
            page_number = int(page_from_url)
        except ValueError:
            if page_from_url == "last":
                page_number = paginator.num_pages
            else:
                raise Http404(_("Page is not “last”, nor can it be converted to an int."))
        try:
            page = paginator.page(page_number)
        except InvalidPage as e:
            raise Http404(
                _("Invalid page (%(page_number)s): %(message)s") % {"page_number": page_number, "message": str(e)}
            )
        object_list = list(page.object_list)
        pagination_context = {
            "paginator": paginator,
            "page_obj": page,
            "is_paginated": page.has_other_pages(),
            "object_list": object_list,
            "renditions_for_posts": list(Post.get_all_renditions_from_queryset(object_list)),
            "page_range": page.paginator.get_elided_page_range(page.number, on_each_side=2, on_ends=1),  # type: ignore
        }
        context.update(pagination_context)
        return context

    @staticmethod
    def get_other_get_params(get_params: QueryDict) -> str:
        filtered_get_params = {k: str(v) for k, v in get_params.items() if k != "page"}
        new_get_params = QueryDict("", mutable=True)
        new_get_params.update(filtered_get_params)
        parameters = new_get_params.urlencode()
        if len(parameters) > 0:
            parameters = f"&{parameters}"
        return parameters

    @property
    def wagtail_api_pages_url(self) -> str:
        return reverse("cast:api:wagtail:pages:listing")

    @property
    def facet_counts_api_url(self) -> str:
        return reverse("cast:api:facet-counts-detail", kwargs={"pk": self.pk})

    @property
    def theme_list_api_url(self) -> str:
        return reverse("cast:api:theme-list")

    @property
    def theme_update_api_url(self) -> str:
        return reverse("cast:api:theme-update")

    @property
    def comment_post_url(self) -> str:
        ajax_post_url = reverse("comments-post-comment-ajax")
        return ajax_post_url

    @property
    def pagination_page_size(self) -> int:
        return appsettings.POST_LIST_PAGINATION

    def get_theme_form(self, request: HtmxHttpRequest) -> django.forms.forms.Form:
        from ..forms import SelectThemeForm

        return SelectThemeForm(
            initial={
                "template_base_dir": self.get_template_base_dir(request),
                "next": request.path,
            }
        )

    def get_context(self, request: HtmxHttpRequest, *args, **kwargs) -> ContextDict:
        context = super().get_context(request, *args, **kwargs)
        get_params = request.GET.copy()
        context["filterset"] = filterset = self.get_filterset(get_params)
        context["parameters"] = self.get_other_get_params(get_params)
        context = self.paginate_queryset(context, self.get_published_posts(filterset.qs), get_params)
        context["posts"] = context["object_list"]  # convenience
        context["blog"] = self
        context["has_selectable_themes"] = True
        context["template_base_dir"] = self.get_template_base_dir(request)
        context["use_audio_player"] = any([post.has_audio for post in context["posts"]])
        context["theme_form"] = self.get_theme_form(request)
        return context


class Podcast(Blog):
    """A podcast is a blog with some extra fields for podcasting."""

    # atm it's only used for podcast image
    itunes_artwork = models.ForeignKey(ItunesArtWork, null=True, blank=True, on_delete=models.SET_NULL)
    itunes_categories = models.CharField(
        _("itunes_categories"),
        max_length=512,
        blank=True,
        default="",
        help_text=_(
            "A json dict of itunes categories pointing to lists "
            "of subcategories. Taken from this list "
            "https://validator.w3.org/feed/docs/error/InvalidItunesCategory.html"
        ),
    )
    keywords = models.CharField(
        _("keywords"),
        max_length=255,
        blank=True,
        default="",
        help_text=_(
            """A comma-delimited list of up to 12 words for iTunes
            searches. Perhaps include misspellings of the title."""
        ),
    )
    EXPLICIT_CHOICES = ((1, _("yes")), (2, _("no")), (3, _("clean")))
    explicit = models.PositiveSmallIntegerField(
        _("explicit"),
        default=1,
        choices=EXPLICIT_CHOICES,
        help_text=_("``Clean`` will put the clean iTunes graphic by it."),
    )

    subpage_types = ["cast.Post", "cast.Episode"]
    is_podcast = True

    objects: PageManager = PageManager()
    aliases_homepage: Any  # don't know why this is needed FIXME

    @property
    def itunes_categories_parsed(self) -> dict[str, list[str]]:
        try:
            return json.loads(self.itunes_categories)
        except json.decoder.JSONDecodeError:
            return {}



from argparse import RawTextHelpFormatter

from django.core.management.base import BaseCommand
from rich.progress import track
from wagtail.images.models import Image, Rendition

from ...models import Blog, Post


class Command(BaseCommand):
    help = """
What does it mean to sync the renditions for all posts?
    - create missing renditions
    - delete obsolete renditions

Optional arguments:
    --post-slug: sync renditions for a specific post
    --blog-slug: sync renditions for posts in a specific blog

By default all posts are synced.
    """

    def create_parser(self, *args, **kwargs):
        parser = super().create_parser(*args, **kwargs)
        parser.formatter_class = RawTextHelpFormatter
        return parser

    def add_arguments(self, parser):
        # Optional argument for a post slug
        parser.add_argument("--post-slug", type=str, help="Sync renditions for a specific post")
        # Optional argument for a blog slug
        parser.add_argument("--blog-slug", type=str, help="Sync renditions for posts in a specific blog")

    def handle(self, *args, **options):
        post_slug = options.get("post_slug")
        blog_slug = options.get("blog_slug")

        if post_slug is not None:
            posts_queryset = Post.objects.filter(slug=post_slug)
        elif blog_slug is not None:
            blog = Blog.objects.get(slug=blog_slug)
            posts_queryset = Post.objects.descendant_of(blog)
        else:
            posts_queryset = Post.objects.all()

        posts_queryset = posts_queryset.prefetch_related("images", "galleries__images")
        all_images = Post.get_all_images_from_queryset(posts_queryset)
        obsolete_renditions, missing_renditions = Post.get_obsolete_and_missing_rendition_strings(all_images)
        Rendition.objects.filter(id__in=obsolete_renditions).delete()
        missing_renditions = list(missing_renditions.items())
        print("len missing renditions: ", len(missing_renditions))
        for image_id, filter_specs in track(missing_renditions, description="create missing renditions"):
            image = Image.objects.get(id=image_id)
            image.get_renditions(*filter_specs)



# Generated by Django 4.1.7 on 2023-03-18 07:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0044_alter_blog_template_base_dir_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blog",
            name="template_base_dir",
            field=models.CharField(
                blank=True,
                choices=[("bootstrap4", "Bootstrap 4"), ("plain", "Just HTML")],
                default=None,
                help_text="The theme to use for this blog implemented as a template base directory. If not set, the template base directory will be determined by a site setting.",
                max_length=128,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="templatebasedirectory",
            name="name",
            field=models.CharField(
                choices=[("bootstrap4", "Bootstrap 4"), ("plain", "Just HTML")],
                default="bootstrap4",
                help_text="The theme to use for this site implemented as a template base directory. It's possible to overwrite this setting for each blog.If you want to use a custom theme, you have to create a new directory in your template directory named cast/<your-theme-name>/ and put all required templates in there.",
                max_length=128,
            ),
        ),
    ]



from collections.abc import Iterator
from contextlib import contextmanager
from io import BytesIO

import django
import pytest

try:
    from django.core.files.storage import storages
except ImportError:
    pass
from django.core.management import CommandError, call_command

try:
    from cast.management.commands.media_backup import Command as MediaBackupCommand
except ImportError:
    pass


def get_comparable_django_version():
    return int("".join(django.get_version().split(".")[:2]))


pytestmark = pytest.mark.skipif(
    get_comparable_django_version() < 42,
    reason="Django version >= 4.2 is required",
)


def test_media_backup_without_storages(settings):
    settings.STORAGES = {}
    with pytest.raises(CommandError) as err:
        call_command("media_backup")
    assert str(err.value) == "production or backup storage not configured"


def test_media_backup_with_wrong_django_version(mocker):
    mocker.patch("cast.management.commands.storage_backend.DJANGO_VERSION_VALID", False)
    with pytest.raises(CommandError) as err:
        call_command("media_backup")
    assert str(err.value) == "Django version >= 4.2 is required"


class StubStorage:
    def __init__(self) -> None:
        self._files: dict[str, BytesIO] = {}
        self._added: set[str] = set()

    def exists(self, path: str) -> bool:
        return path in self._files

    def was_added_by_backup(self, name: str) -> bool:
        return name in self._added

    def was_not_added_by_backup(self, name: str) -> bool:
        return name not in self._added

    def save(self, name: str, content: BytesIO) -> None:
        self.save_without_adding(name, content)
        self._added.add(name)

    def save_without_adding(self, name: str, content: BytesIO) -> None:
        self._files[name] = content

    def listdir(self, _path: str) -> tuple[list, dict[str, BytesIO]]:
        return [], self._files

    @contextmanager
    def open(self, name: str, _mode: str) -> Iterator[BytesIO]:
        try:
            yield self._files[name]
        finally:
            pass


@pytest.fixture
def stub_storages(settings):
    storage_stub = {"BACKEND": "tests.management_command_test.StubStorage"}
    settings.STORAGES = {"production": storage_stub, "backup": storage_stub}
    return storages


def test_media_backup_new_file_in_production(stub_storages):
    production, backup = stub_storages["production"], stub_storages["backup"]

    # given there's a new file added to production
    production.save_without_adding("foobar.jpg", BytesIO(b"foobar"))  # type: ignore

    # when we run the backup command
    call_command("media_backup")

    # then the file should have been added by the backup command
    assert backup.was_added_by_backup("foobar.jpg")  # type: ignore


def test_media_backup_existing_file_in_backup(stub_storages):
    production, backup = stub_storages["production"], stub_storages["backup"]

    # given there's a file in the backup
    production.save_without_adding("foobar.jpg", BytesIO(b"foobar"))  # type: ignore
    backup.save_without_adding("foobar.jpg", BytesIO(b"foobar"))  # type: ignore

    # when we run the backup method
    MediaBackupCommand().backup_media_files(production, backup)

    # then the file should not have been added by the backup command
    assert backup.was_not_added_by_backup("foobar.jpg")  # type: ignore



import pytest
from django.urls import reverse


@pytest.mark.django_db
def test_twitter_player_blog_does_not_match(client, blog, episode):
    assert blog != episode.blog
    url = reverse("cast:twitter-player", kwargs={"blog_slug": blog.slug, "episode_slug": episode.slug})
    r = client.get(url)
    assert r.status_code == 404



# Generated by Django 3.2.5 on 2021-07-12 09:19

import django.db.models.deletion
import taggit.managers
import wagtail.blocks
import wagtail.fields
import wagtail.models
import wagtail.embeds.blocks
import wagtail.images.blocks
from django.db import migrations, models

import cast.blocks


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0062_comment_models_and_pagesubscription"),
        ("taggit", "0003_taggeditem_add_unique_index"),
        ("cast", "0007_alter_post_body"),
    ]

    operations = [
        migrations.AddField(
            model_name="video",
            name="collection",
            field=models.ForeignKey(
                default=wagtail.models.get_root_collection_id,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="wagtailcore.collection",
                verbose_name="collection",
            ),
        ),
        migrations.AddField(
            model_name="video",
            name="tags",
            field=taggit.managers.TaggableManager(
                blank=True, help_text=None, through="taggit.TaggedItem", to="taggit.Tag", verbose_name="tags"
            ),
        ),
        migrations.AddField(
            model_name="video",
            name="title",
            field=models.CharField(default="", max_length=255),
        ),
        migrations.AlterField(
            model_name="post",
            name="body",
            field=wagtail.fields.StreamField(
                [
                    ("heading", wagtail.blocks.CharBlock(form_classname="full title")),
                    ("paragraph", wagtail.blocks.RichTextBlock()),
                    ("image", wagtail.images.blocks.ImageChooserBlock(template="cast/image.html")),
                    ("gallery", cast.blocks.GalleryBlock(wagtail.images.blocks.ImageChooserBlock())),
                    ("embed", wagtail.embeds.blocks.EmbedBlock()),
                    ("video", cast.blocks.VideoChooserBlock(template="cast/video/video.html")),
                ]
            ),
        ),
    ]



import factory
from django.contrib.auth import get_user_model
from wagtail.models import Site

from cast.models import Blog, Episode, Gallery, Podcast, Post, Video


class SiteFactory(factory.django.DjangoModelFactory):
    hostname = "localhost"

    class Meta:
        model = Site
        django_get_or_create = ("hostname",)


class UserFactory(factory.django.DjangoModelFactory):
    username = factory.Sequence(lambda n: f"user-{n}")
    email = factory.Sequence(lambda n: f"user-{n}@example.com")
    password = factory.PostGenerationMethodCall("set_password", "password")

    class Meta:
        model = get_user_model()
        django_get_or_create = ("username",)


class VideoFactory(factory.django.DjangoModelFactory):
    user = None
    original = factory.django.ImageField(color="blue")

    class Meta:
        model = Video


class GalleryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Gallery


class PageFactory(factory.django.DjangoModelFactory):
    class Meta:
        abstract = True

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        parent = kwargs.pop("parent")
        page = model_class(*args, **kwargs)
        parent.add_child(instance=page)
        return page


class BlogFactory(PageFactory):
    author = None
    title = factory.Sequence(lambda n: f"blog-{n}")
    slug = factory.Sequence(lambda n: f"blog-{n}")

    class Meta:
        model = Blog
        django_get_or_create = ("slug",)


class PodcastFactory(PageFactory):
    author = None
    title = factory.Sequence(lambda n: f"blog-{n}")
    slug = factory.Sequence(lambda n: f"blog-{n}")

    class Meta:
        model = Podcast
        django_get_or_create = ("slug",)


class PostFactory(PageFactory):
    class Meta:
        model = Post
        django_get_or_create = ("slug",)


class EpisodeFactory(PageFactory):
    class Meta:
        model = Episode
        django_get_or_create = ("slug",)



import logging
import uuid
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any, Optional, Union

from django import forms
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.safestring import SafeText
from django.utils.translation import gettext_lazy as _
from django_comments import get_model as get_comment_model
from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from rest_framework.fields import Field
from slugify import slugify
from taggit.models import TaggedItemBase
from wagtail import blocks
from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.admin.panels import FieldPanel, MultiFieldPanel
from wagtail.api import APIField
from wagtail.embeds.blocks import EmbedBlock
from wagtail.fields import StreamField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image, Rendition
from wagtail.models import Page, PageManager
from wagtail.search import index

from cast import appsettings
from cast.blocks import (
    AudioChooserBlock,
    CastImageChooserBlock,
    CodeBlock,
    GalleryBlock,
    GalleryBlockWithLayout,
    VideoChooserBlock,
)
from cast.models import get_or_create_gallery

from ..renditions import ImageType, RenditionFilters
from .theme import TemplateBaseDirectory

if TYPE_CHECKING:
    from .index_pages import Blog, ContextDict, Podcast

logger = logging.getLogger(__name__)


comment_model = get_comment_model()


class ContentBlock(blocks.StreamBlock):
    heading: blocks.CharBlock = blocks.CharBlock(classname="full title")
    paragraph: blocks.RichTextBlock = blocks.RichTextBlock()
    code: CodeBlock = CodeBlock(icon="code")
    image: CastImageChooserBlock = CastImageChooserBlock(template="cast/image/image.html")
    gallery: GalleryBlockWithLayout = GalleryBlockWithLayout()
    embed: EmbedBlock = EmbedBlock()
    video: VideoChooserBlock = VideoChooserBlock(template="cast/video/video.html", icon="media")
    audio: AudioChooserBlock = AudioChooserBlock(template="cast/audio/audio.html", icon="media")

    class Meta:
        icon = "form"


TypeToIdSet = dict[str, set[int]]


def sync_media_ids(from_database: TypeToIdSet, from_body: TypeToIdSet) -> tuple[TypeToIdSet, TypeToIdSet]:
    to_add, to_remove = {}, {}
    all_media_types = set(from_database.keys()).union(from_body.keys())
    for media_type in all_media_types:
        in_database_ids = from_database.get(media_type, set())
        in_body_ids = from_body.get(media_type, set())
        ids_to_add = in_body_ids - in_database_ids
        if len(ids_to_add) > 0:
            to_add[media_type] = ids_to_add
        ids_to_remove = in_database_ids - in_body_ids
        if len(ids_to_remove) > 0:
            to_remove[media_type] = ids_to_remove
    return to_add, to_remove


class PostTag(TaggedItemBase):
    content_object = ParentalKey("Post", related_name="tagged_items", on_delete=models.CASCADE)


class HtmlField(Field):
    """
    A serializer field to render the html of a post. It's used in the wagtail api
    to get the rendered html of the overview and detail of a post. An SPA theme
    like `cast_vue` can then use this html instead of having to render the blocks
    itself.
    """

    def __init__(self, *, render_detail: bool = False, **kwargs) -> None:
        self.render_detail = render_detail
        super().__init__(**kwargs)

    def to_representation(self, post: "Post") -> SafeText:
        """
        Pass the request from context to the post's serve method to be able to
        render the post with the correct theme.
        """
        return post.get_description(
            request=self.context["request"], render_detail=self.render_detail, escape_html=False, remove_newlines=False
        )


ImagesWithType = Iterator[tuple[ImageType, Image]]


class Post(Page):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    visible_date = models.DateTimeField(
        default=timezone.now,
        help_text=_("The visible date of the post which is used for sorting."),
        db_index=True,
    )
    comments_enabled = models.BooleanField(
        default=True,
        help_text=_("Whether comments are enabled for this post."),
    )

    images: models.ManyToManyField = models.ManyToManyField(Image, blank=True)  # FIXME mypy are you ok?
    videos: models.ManyToManyField = models.ManyToManyField("cast.Video", blank=True)
    galleries: models.ManyToManyField = models.ManyToManyField("cast.Gallery", blank=True)
    audios: models.ManyToManyField = models.ManyToManyField("cast.Audio", blank=True)
    categories = ParentalManyToManyField("cast.PostCategory", blank=True)

    # managers
    tags = ClusterTaggableManager(through=PostTag, blank=True, verbose_name=_("tags"))

    _local_template_name: Optional[str] = None

    # wagtail
    body = StreamField(
        [
            ("overview", ContentBlock()),
            ("detail", ContentBlock()),
        ],
        use_json_field=True,
    )

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]

    api_fields = [
        APIField("uuid"),
        APIField("visible_date"),
        APIField("comments_are_enabled"),
        APIField("body"),
        APIField("html_overview", serializer=HtmlField(source="*", render_detail=False)),
        APIField("html_detail", serializer=HtmlField(source="*", render_detail=True)),
        APIField("comments"),
        APIField("comments_security_data"),
        APIField("podlove_players"),
    ]

    content_panels = Page.content_panels + [
        FieldPanel("visible_date"),
        FieldPanel("categories", widget=forms.CheckboxSelectMultiple),
        FieldPanel("tags"),
        FieldPanel("body"),
    ]
    parent_page_types = ["cast.Blog", "cast.Podcast"]

    @property
    def media_model_lookup(self) -> dict[str, type[models.Model]]:
        from .audio import Audio
        from .gallery import Gallery
        from .video import Video

        return {
            "image": Image,
            "video": Video,
            "gallery": Gallery,
            "audio": Audio,
        }

    @property
    def blog(self) -> "Blog":
        """
        The get_parent() method returns wagtail parent page, which is not
        necessarily a Blog model, but maybe the root page. If it's a Blog
        it has a .blog attribute containing the model which has all the
        attributes like blog.comments_enabled etc.
        """
        return self.get_parent().blog

    def get_template_base_dir(self, request: HttpRequest) -> str:
        parent = self.get_parent()
        if parent is not None:
            return parent.blog.get_template_base_dir(request)
        else:
            return TemplateBaseDirectory.for_request(request).name

    def get_template(self, request: HttpRequest, *args, local_template_name: str = "post.html", **kwargs) -> str:
        if self._local_template_name is not None:
            local_template_name = self._local_template_name
        template_base_dir = self.get_template_base_dir(request)
        template = f"cast/{template_base_dir}/{local_template_name}"
        return template

    def __str__(self) -> str:
        return self.title

    def get_slug(self) -> str:
        return slugify(self.title)

    @property
    def media_lookup(self) -> dict[str, dict[int, Any]]:
        try:
            return {
                "image": {i.pk: i for i in self.images.all()},
                "video": {v.pk: v for v in self.videos.all()},
                "gallery": {g.pk: g for g in self.galleries.all()},
                "audio": {a.pk: a for a in self.audios.all()},
            }
        except ValueError:
            # post ist not yet saved
            pass
        return {}

    @property
    def media_attr_lookup(self) -> "ContextDict":
        return {
            "image": self.images,
            "video": self.videos,
            "gallery": self.galleries,
            "audio": self.audios,
        }

    @property
    def audio_in_body(self) -> bool:
        audio_blocks = []
        for block in self.body:
            for content_block in block.value:
                if content_block.block_type == "audio":
                    audio_blocks.append(content_block)
        return len(audio_blocks) > 0

    @property
    def has_audio(self) -> bool:
        audios_count = 0
        try:
            audios_count = self.audios.count()
        except ValueError:
            # will be raised on wagtail preview because page_ptr is not set
            pass
        if self.audio_in_body or audios_count > 0:
            return True
        else:
            if isinstance(self.specific, Episode):
                return self.specific.podcast_audio is not None
            else:
                return False

    @property
    def comments_are_enabled(self) -> bool:
        return appsettings.CAST_COMMENTS_ENABLED and self.blog.comments_enabled and self.comments_enabled

    @property
    def comments_security_data(self) -> dict[str, Union[str, int]]:
        from django_comments.forms import CommentSecurityForm

        form = CommentSecurityForm(self)
        return form.generate_security_data()

    @property
    def page_type(self) -> str:
        """
        cannot use wagtail.api.v2.serializers.TypeField.to_representation easily
        """
        return type(self)._meta.app_label + "." + type(self).__name__  # FIXME butt ugly

    def get_context(self, request: HttpRequest, **kwargs) -> "ContextDict":
        context = super().get_context(request, **kwargs)
        context["render_detail"] = kwargs.get("render_detail", False)
        # needed for blocks with themed templates
        context["template_base_dir"] = self.get_template_base_dir(request)
        context["blog"] = self.blog  # needed for SPA themes
        return context

    @property
    def media_ids_from_db(self) -> TypeToIdSet:
        return {k: set(v) for k, v in self.media_lookup.items()}

    def _media_ids_from_body(self, body: StreamField) -> TypeToIdSet:
        from_body: TypeToIdSet = {}
        for content_block in body:
            for block in content_block.value:
                if block.block_type == "gallery":
                    images = block.value.get("gallery", [])
                    image_ids = [i.id for i in images]
                    media_model = get_or_create_gallery(image_ids)
                else:
                    media_model = block.value
                if block.block_type in self.media_model_lookup:
                    if media_model is not None:
                        from_body.setdefault(block.block_type, set()).add(media_model.id)
        return from_body

    @property
    def media_ids_from_body(self) -> TypeToIdSet:
        return self._media_ids_from_body(self.body)

    def sync_media_ids(self) -> None:
        media_attr_lookup = self.media_attr_lookup
        to_add, to_remove = sync_media_ids(self.media_ids_from_db, self.media_ids_from_body)

        # add new ids
        for media_type, ids in to_add.items():
            for media_id in ids:
                media_attr_lookup[media_type].add(media_id)

        # remove obsolete ids
        for media_type, ids in to_remove.items():
            for media_id in ids:
                media_attr_lookup[media_type].remove(media_id)

    # helper methods for image rendition syncing
    @staticmethod
    def get_all_images_from_queryset(posts: Iterable["Post"]) -> ImagesWithType:
        for post in posts:
            yield from post.get_all_images()

    @staticmethod
    def get_all_renditions_from_queryset(posts: Iterable["Post"]) -> Iterator[Rendition]:
        for image_type, image in Post.get_all_images_from_queryset(posts):
            yield from image.renditions.all()

    @staticmethod
    def get_all_filterstrings(images_with_type: ImagesWithType) -> Iterator[tuple[int, str]]:
        for image_type, image in images_with_type:
            rfs = RenditionFilters.from_wagtail_image_with_type(image, image_type)
            for filter_string in rfs.filter_strings:
                yield image.pk, filter_string

    def get_all_images(self) -> ImagesWithType:
        """
        Use it like this:
        posts_queryset = Post.objects.prefetch_related("images", "galleries__images")[:10]
        for post in posts_queryset:
            for image_type, image in post.get_all_images():
                print(image_type, image)
        """
        for image in self.images.all():
            yield "regular", image
        for gallery in self.galleries.all():
            for image in gallery.images.all():
                yield "gallery", image

    @staticmethod
    def get_obsolete_and_missing_rendition_strings(
        images_with_type: ImagesWithType,
    ) -> tuple[set[int], dict[int, set[str]]]:
        """
        Get all obsolete and missing rendition strings from a queryset of posts.
        """
        required_renditions = set(Post.get_all_filterstrings(images_with_type))
        all_image_ids = {image_id for image_id, filter_string in required_renditions}
        renditions_queryset = Rendition.objects.filter(image__in=all_image_ids)
        existing_rendition_to_id = {
            (image_id, filter_spec): pk
            for pk, image_id, filter_spec in renditions_queryset.values_list("pk", "image_id", "filter_spec")
        }
        existing_renditions = set(existing_rendition_to_id.keys())
        obsolete_renditions_unfiltered = existing_renditions - required_renditions

        # remove wagtail generated renditions from obsolete_renditions
        obsolete_renditions = set()
        wagtail_filter_specs = {"max-165x165", "max-800x600"}
        for image_id, filter_spec in obsolete_renditions_unfiltered:
            if filter_spec not in wagtail_filter_specs:
                obsolete_renditions.add((image_id, filter_spec))
        obsolete_rendition_pks = {
            existing_rendition_to_id[(image_id, filter_spec)] for image_id, filter_spec in obsolete_renditions
        }

        # build missing renditions aggregated by image id
        missing_renditions = required_renditions - existing_renditions
        missing_renditions_by_image_id: dict[int, set[str]] = {}  # why mypy?
        for image_id, filter_spec in missing_renditions:
            missing_renditions_by_image_id.setdefault(image_id, set()).add(filter_spec)
        return obsolete_rendition_pks, missing_renditions_by_image_id

    @property
    def comments(self) -> list[dict[str, Union[int, None, str]]]:
        ctype = ContentType.objects.get_for_model(self)
        site_id = getattr(settings, "SITE_ID", None)
        qs = comment_model.objects.filter(
            content_type=ctype,
            object_pk=smart_str(self.pk),
            site__pk=site_id,
        )

        # The is_public and is_removed fields are implementation details of the
        # built-in comment model's spam filtering system, so they might not
        # be present on a custom comment model subclass. If they exist, we
        # should filter on them.
        field_names = [f.name for f in comment_model._meta.fields]
        if "is_public" in field_names:
            qs = qs.filter(is_public=True)
        if getattr(settings, "COMMENTS_HIDE_REMOVED", True) and "is_removed" in field_names:
            qs = qs.filter(is_removed=False)

        result = []
        for comment in qs:
            result.append(
                {
                    "id": comment.id,
                    "parent": comment.parent_id,
                    "user": comment.user_name,
                    "date": comment.submit_date,
                    "comment": comment.comment,
                }
            )
        return result

    @property
    def podlove_players(self) -> list[tuple[str, str]]:
        """
        Get the podlove player data for posts containing audio elements.
        """
        result = []
        for pk, audio in self.media_lookup["audio"].items():
            element_id = f"#audio_{pk}"
            result.append((element_id, audio.podlove_url))
        return result

    def get_description(
        self,
        *,
        request: HttpRequest,
        render_detail: bool = False,
        escape_html: bool = True,
        remove_newlines: bool = True,
    ) -> SafeText:
        """
        Get a description for the feed or twitter player card. Needs to be
        a method because the feed is able to pass the actual request object.
        """
        self._local_template_name = "post_body.html"
        description = self.serve(request, render_detail=render_detail).rendered_content
        if remove_newlines:
            description = description.replace("\n", "")
        if escape_html:
            description = escape(description)
        return description

    def get_absolute_url(self) -> str:
        """This is needed for django-fluentcomments."""
        return self.full_url

    def save(self, *args, **kwargs) -> None:
        save_return = super().save(*args, **kwargs)
        self.sync_media_ids()
        return save_return


class CustomEpisodeForm(WagtailAdminPageForm):
    """
    Custom form for Episode to validate the podcast_audio field.

    The reason for this is that the podcast_audio field is not required
    for draft episodes, but it is required for published episodes. So
    we have to check which button was clicked in the admin form.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["action-publish"] = forms.CharField(required=False, widget=forms.HiddenInput())

    def clean(self) -> dict[str, Any]:
        cleaned_data = super().clean()
        if cleaned_data.get("action-publish") and cleaned_data.get("podcast_audio") is None:
            raise forms.ValidationError({"podcast_audio": _("An episode must have an audio file to be published.")})
        return cleaned_data


class Episode(Post):
    """A podcast episode is just a Post with some additional fields."""

    podcast_audio = models.ForeignKey(
        "cast.Audio",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="episodes",
        help_text=_(
            "The audio file for this episode -if this is not set, the episode will not be included in the feed."
        ),
    )
    keywords = models.CharField(
        max_length=255,
        blank=True,
        default="",
        help_text=_(
            """A comma-delimited-list of up to 12 words for iTunes
            searches. Perhaps include misspellings of the title."""
        ),
    )
    EXPLICIT_CHOICES = ((1, _("yes")), (2, _("no")), (3, _("clean")))
    explicit = models.PositiveSmallIntegerField(
        choices=EXPLICIT_CHOICES,
        help_text=_("``Clean`` will put the clean iTunes graphic by it."),
        default=1,
    )
    block = models.BooleanField(
        default=False,
        help_text=_(
            "Check to block this episode from iTunes because <br />its "
            "content might cause the entire show to be <br />removed from iTunes."
            ""
        ),
    )

    parent_page_types = ["cast.Podcast"]

    content_panels = Page.content_panels + [
        FieldPanel("visible_date"),
        FieldPanel("podcast_audio"),
        MultiFieldPanel(
            [FieldPanel("categories", widget=forms.CheckboxSelectMultiple)],
            heading="Categories",
            classname="collapsed",
        ),
        FieldPanel("tags"),
        FieldPanel("body"),
        FieldPanel("keywords"),
        FieldPanel("explicit"),
        FieldPanel("block"),
    ]

    objects: PageManager = PageManager()
    aliases_homepage: Any  # FIXME: why is this needed?
    base_form_class = CustomEpisodeForm

    def get_template(self, request: HttpRequest, *args, **kwargs) -> str:
        """
        Use get_template() from the parent class, but pass a local template name.
        """
        return super().get_template(request, *args, local_template_name="episode.html", **kwargs)

    def get_context(self, request, **kwargs) -> "ContextDict":
        context = super().get_context(request, **kwargs)
        context["episode"] = self
        if hasattr(request, "build_absolute_uri"):
            player_url = reverse(
                "cast:twitter-player", kwargs={"episode_slug": self.slug, "blog_slug": self.podcast.slug}
            )
            player_url = request.build_absolute_uri(player_url)
            context["player_url"] = player_url
        return context

    @property
    def podcast(self) -> "Podcast":
        """
        The get_parent() method returns wagtail parent page, which is not
        necessarily a Blog model, but maybe the root page. If it's a Blog
        it has a .blog attribute containing the model which has all the
        attributes like blog.comments_enabled etc.
        """
        return self.get_parent().specific

    def get_enclosure_url(self, audio_format: str) -> str:
        return getattr(self.podcast_audio, audio_format).url

    def get_enclosure_size(self, audio_format: str) -> int:
        if self.podcast_audio is None:
            return 0
        return self.podcast_audio.get_file_size(audio_format)


class HomePage(Page):
    body = StreamField(
        [
            ("heading", blocks.CharBlock(classname="full title")),
            ("paragraph", blocks.RichTextBlock()),
            ("image", CastImageChooserBlock(template="cast/image/image.html")),
            ("gallery", GalleryBlock(ImageChooserBlock())),
        ],
        use_json_field=True,
    )
    alias_for_page = models.ForeignKey(
        Page,
        related_name="aliases_homepage",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        verbose_name="Redirect to another page",
        help_text="Make this page an alias for another page, redirecting to it with a non permanent redirect.",
    )

    content_panels = Page.content_panels + [
        FieldPanel("alias_for_page"),
        FieldPanel("body"),
    ]

    def serve(self, request, *args, **kwargs) -> HttpResponse:
        if self.alias_for_page is not None:
            return redirect(self.alias_for_page.url, permanent=False)
        return super().serve(request, *args, **kwargs)



# Generated by Django 4.1.7 on 2023-03-05 21:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0039_blog_noindex"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blog",
            name="noindex",
            field=models.BooleanField(
                default=False,
                help_text="Whether to add a noindex meta tag to this page and all subpages excluding them from search engines.",
                verbose_name="noindex",
            ),
        ),
    ]



from collections.abc import Iterable
from itertools import chain, islice, tee
from typing import TYPE_CHECKING, Optional, Union

from django.db.models import QuerySet
from django.template.loader import TemplateDoesNotExist, get_template
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import ClassNotFound, get_lexer_by_name
from wagtail.blocks import CharBlock, ChoiceBlock, ListBlock, StructBlock, TextBlock
from wagtail.images.blocks import ChooserBlock, ImageChooserBlock
from wagtail.images.models import AbstractImage, AbstractRendition

from . import appsettings as settings
from .renditions import (
    Height,
    ImageForSlot,
    ImageType,
    Rectangle,
    RenditionFilters,
    Width,
)

if TYPE_CHECKING:
    from .models import Audio, Video
    from .widgets import AdminVideoChooser


def previous_and_next(all_items: Iterable) -> Iterable:
    """
    Turn an iterable into an iterable of tuples of the
    previous and next item in the iterable.

    Example:
        >>> list(previous_and_next(range(3)))
        [(None, 0, 1), (0, 1, 2), (1, 2, None)]
    """
    previous_items, items, next_items = tee(all_items, 3)
    previous_items = chain([None], previous_items)
    next_items = chain(islice(next_items, 1, None), [None])
    return zip(previous_items, items, next_items)


def get_srcset_images_for_slots(
    image: AbstractImage,
    image_type: ImageType,
    fetched_renditions: dict[str, AbstractRendition] | None = None,
) -> dict[Rectangle, ImageForSlot]:
    """
    Get the srcset images for the given slots and image formats. This will fetch
    renditions from wagtail and return a list of ImageInSlot objects.
    """
    images_for_slots = {}
    rendition_filters = RenditionFilters.from_wagtail_image_with_type(image=image, image_type=image_type)
    slots, image_formats = rendition_filters.slots, rendition_filters.image_formats
    rendition_filter_strings = rendition_filters.filter_strings
    if len(rendition_filter_strings) > 0:
        renditions = {}
        if fetched_renditions is not None:
            renditions = fetched_renditions
        filter_strings_to_fetch = [fs for fs in rendition_filter_strings if fs not in renditions]
        if len(filter_strings_to_fetch) > 0:
            renditions.update(image.get_renditions(*filter_strings_to_fetch))
        rendition_filters.set_filter_to_url_via_wagtail_renditions(renditions)
    for slot in slots:
        try:
            images_for_slots[slot] = rendition_filters.get_image_for_slot(slot)
        except ValueError:
            # no fitting image found for slot -> use original image
            src = {}
            for image_format in image_formats:
                if image_format == rendition_filters.original_format:
                    src[image_format] = image.file.url
                else:
                    # convert to image_format
                    rendition = image.get_rendition(f"format-{image_format}")
                    src[image_format] = rendition.url
            srcset = {}
            for image_format in image_formats:
                if image_format == rendition_filters.original_format:
                    srcset[image_format] = f"{image.file.url} {image.width}w"
                else:
                    # convert to image_format
                    rendition = image.get_rendition(f"format-{image_format}")
                    srcset[image_format] = f"{rendition.url} {rendition.width}w"
            width = rendition_filters.slot_to_fitting_width[slot]
            images_for_slots[slot] = ImageForSlot(Rectangle(width, slot.height), src, srcset)
    return images_for_slots


class CastImageChooserBlock(ImageChooserBlock):
    """
    Just add a thumbnail to the image because we then can use the thumbnail
    to get the srcset and sizes attributes in the template.
    """

    def get_context(self, image: AbstractImage, parent_context: Optional[dict] = None) -> dict:
        if parent_context is None:
            parent_context = {}
        fetched_renditions = {
            r.filter_spec: r for r in parent_context.get("renditions_for_posts", {}) if r.image_id == image.pk
        }
        images_for_slots = get_srcset_images_for_slots(image, "regular", fetched_renditions=fetched_renditions)
        [image.regular] = images_for_slots.values()
        return super().get_context(image, parent_context=parent_context)


def add_prev_next(images: QuerySet[AbstractImage]) -> None:
    """
    For each image in the queryset, add the previous and next image.
    """
    for previous_image, current_image, next_image in previous_and_next(images):
        current_image.prev = "false" if previous_image is None else f"img-{previous_image.pk}"
        current_image.next = "false" if next_image is None else f"img-{next_image.pk}"


def add_image_thumbnails(images: QuerySet[AbstractImage], context: dict) -> None:
    """
    For each image in the queryset, add the thumbnail and modal image data to the image.
    """
    modal_slot, thumbnail_slot = (
        Rectangle(Width(w), Height(h)) for w, h in settings.CAST_GALLERY_IMAGE_SLOT_DIMENSIONS
    )
    for image in images:
        fetched_renditions = {
            r.filter_spec: r for r in context.get("renditions_for_posts", {}) if r.image_id == image.pk
        }
        images_for_slots = get_srcset_images_for_slots(image, "gallery", fetched_renditions=fetched_renditions)
        image.modal = images_for_slots[modal_slot]
        image.thumbnail = images_for_slots[thumbnail_slot]


def prepare_context_for_gallery(images: QuerySet[AbstractImage], context: dict) -> dict:
    """
    Add the previous and next image and the thumbnail and modal image data to each
    image of the gallery and then the images to the context.
    """
    add_prev_next(images)
    add_image_thumbnails(images, context=context)
    context["image_pks"] = ",".join([str(image.pk) for image in images])
    context["images"] = images
    return context


def get_gallery_block_template(default_template_name: str, context: Optional[dict], layout: str = "default") -> str:
    print("default_template_name: ", default_template_name, "context: ", context, "layout: ", layout)
    if context is None:
        return default_template_name

    template_base_dir = context.get("template_base_dir")
    if template_base_dir is None:
        return default_template_name

    if layout == "htmx":
        template_from_theme = f"cast/{template_base_dir}/gallery_htmx.html"
    else:
        template_from_theme = f"cast/{template_base_dir}/gallery.html"
    try:
        get_template(template_from_theme)
        return template_from_theme
    except TemplateDoesNotExist:
        return default_template_name


class GalleryBlock(ListBlock):
    class Meta:
        icon = "image"
        label = "Gallery"
        template = "cast/gallery.html"

    def get_template(self, images: Optional[QuerySet[AbstractImage]] = None, context: Optional[dict] = None) -> str:
        default_template_name = super().get_template(images, context)
        return get_gallery_block_template(default_template_name, context)

    def get_context(self, images: QuerySet[AbstractImage], parent_context: Optional[dict] = None) -> dict:
        context = super().get_context(images, parent_context=parent_context)
        return prepare_context_for_gallery(images, context)


class GalleryBlockWithLayout(StructBlock):
    """
    A gallery block with a layout. The layout parameter controls
    which template is used to render the gallery.
    """

    gallery = GalleryBlock(ImageChooserBlock())
    layout = ChoiceBlock(
        choices=[
            ("default", _("Web Component with Modal")),
            ("htmx", _("HTMX based layout")),
        ],
        default="default",
    )

    class Meta:
        icon = "image"
        label = "Gallery with Layout"
        template = "cast/gallery.html"

    def get_template(self, value=None, context=None):
        default_template_name = super().get_template(value, context)
        layout = "default"
        if value is not None:
            if (layout_from_value := value.get("layout")) is not None:
                layout = layout_from_value
        return get_gallery_block_template(default_template_name, context, layout=layout)

    def get_context(self, value, parent_context: Optional[dict] = None):
        context = super().get_context(value, parent_context=parent_context)
        context["post_pk"] = context["page"].pk
        return prepare_context_for_gallery(value["gallery"], context)


class VideoChooserBlock(ChooserBlock):
    @cached_property
    def target_model(self) -> type["Video"]:
        from .models import Video

        return Video

    @cached_property
    def widget(self) -> "AdminVideoChooser":
        from .widgets import AdminVideoChooser

        return AdminVideoChooser()

    def get_form_state(self, value: Optional[Union["Video", int]]) -> Optional[dict]:
        return self.widget.get_value_data(value)


class AudioChooserBlock(ChooserBlock):
    @cached_property
    def target_model(self) -> type["Audio"]:
        from .models import Audio

        return Audio

    @cached_property
    def widget(self):
        from .widgets import AdminAudioChooser

        return AdminAudioChooser()

    def get_form_state(self, value: Optional[Union["Video", int]]) -> Optional[dict]:
        return self.widget.get_value_data(value)


class CodeBlock(StructBlock):
    language = CharBlock(help_text="The language of the code block")
    source = TextBlock(rows=8, help_text="The source code of the block")

    def render_basic(self, value: Optional[dict], context=None) -> str:
        if value is not None:
            try:
                lexer = get_lexer_by_name(value["language"], stripall=True)
            except (ClassNotFound, KeyError):
                lexer = get_lexer_by_name("text", stripall=True)
            highlighted = highlight(value["source"], lexer, HtmlFormatter())
            return mark_safe(highlighted)
        else:
            return ""



# Generated by Django 4.1.7 on 2023-03-05 21:47

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("wagtailcore", "0083_workflowcontenttype"),
        ("cast", "0040_alter_blog_noindex"),
    ]

    operations = [
        migrations.CreateModel(
            name="TemplateBaseDirectory",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "name",
                    models.CharField(
                        choices=[("bootstrap4", "Bootstrap 4"), ("plain", "Just HTML")],
                        default="bootstrap4",
                        max_length=10,
                    ),
                ),
                (
                    "site",
                    models.OneToOneField(
                        editable=False, on_delete=django.db.models.deletion.CASCADE, to="wagtailcore.site"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from typing import Any

from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.decorators.vary import vary_on_headers
from wagtail.admin import messages
from wagtail.admin.modal_workflow import render_modal_workflow
from wagtail.admin.models import popular_tags_for_model
from wagtail.search.backends import get_search_backends

from ..appsettings import CHOOSER_PAGINATION, MENU_ITEM_PAGINATION
from ..forms import AudioForm, NonEmptySearchForm
from ..models import Audio
from . import AuthenticatedHttpRequest
from .wagtail_pagination import paginate, pagination_template


@vary_on_headers("X-Requested-With")
def index(request: HttpRequest) -> HttpResponse:
    ordering = "-created"
    audios = Audio.objects.all().order_by(ordering)

    # Search
    query_string = None
    if "q" in request.GET:
        form = NonEmptySearchForm(request.GET, placeholder=_("Search audio files"))
        if form.is_valid():
            query_string = form.cleaned_data["q"]
            audios = audios.search(query_string)
    else:
        form = NonEmptySearchForm(placeholder=_("Search media"))

    # Pagination
    paginator, audios = paginate(request, audios, per_page=MENU_ITEM_PAGINATION)

    # Create response
    if request.headers.get("x-requested-with") == "XMLHttpRequest":
        return render(
            request,
            "cast/audio/results.html",
            {
                "ordering": ordering,
                "audios": audios,
                "query_string": query_string,
                "is_searching": bool(query_string),
            },
        )
    else:
        return render(
            request,
            "cast/audio/index.html",
            {
                "ordering": ordering,
                "audios": audios,
                "query_string": query_string,
                "is_searching": bool(query_string),
                "search_form": form,
                "popular_tags": popular_tags_for_model(Audio),
                "user_can_add": True,
                "collections": None,
                "current_collection": None,
            },
        )


def add(request: AuthenticatedHttpRequest) -> HttpResponse:
    if request.POST:
        audio = Audio(user=request.user)
        form = AudioForm(request.POST, request.FILES, instance=audio, user=request.user)
        if form.is_valid():
            form.save()

            # Reindex the media entry to make sure all tags are indexed
            for backend in get_search_backends():
                backend.add(audio)

            messages.success(
                request,
                _("Audio file '{0}' added.").format(audio.title),
                buttons=[messages.button(reverse("castaudio:edit", args=(audio.id,)), _("Edit"))],
            )
            return redirect("castaudio:index")
        else:
            messages.error(request, _("The audio file could not be saved due to errors."))
    else:
        audio = Audio(user=request.user)
        form = AudioForm(user=request.user, instance=audio)

    return render(
        request,
        "cast/audio/add.html",
        {"form": form},
    )


def delete_old_audio_files(audio: Audio, changed_audio_files: set[str]) -> None:
    for file_format in changed_audio_files:
        # if providing a new audio file, delete the old one.
        # NB Doing this via original_file.delete() clears the file field,
        # which definitely isn't what we want...
        original_file = getattr(audio, file_format)
        if original_file.name != "":
            original_file.storage.delete(original_file.name)


def edit(request: HttpRequest, audio_id: int) -> HttpResponse:
    audio = get_object_or_404(Audio, id=audio_id)

    if request.method == "POST":
        form = AudioForm(request.POST, request.FILES, instance=audio, user=request.user)
        if form.is_valid():
            changed_audio_files = set(form.changed_data).intersection(audio.audio_formats)
            if len(changed_audio_files) > 0:
                # FIXME butt ugly
                old_audio = get_object_or_404(Audio, id=audio_id)
                delete_old_audio_files(old_audio, changed_audio_files)
            audio = form.save()

            # Reindex the media entry to make sure all tags are indexed
            for backend in get_search_backends():
                backend.add(audio)

            messages.success(
                request,
                _("Audio file '{0}' updated").format(audio.title),
                buttons=[messages.button(reverse("castaudio:edit", args=(audio.id,)), _("Edit"))],
            )
            return redirect("castaudio:index")
        else:
            messages.error(request, _("The media could not be saved due to errors."))
    else:
        form = AudioForm(instance=audio, user=request.user, initial={"chaptermarks": audio.chapters_as_text})

    filesize = None

    # Get file size when there is a file associated with the Media object
    if audio.m4a:
        try:
            filesize = audio.m4a.size
        except OSError:
            # File doesn't exist
            pass

    if not filesize:
        messages.error(
            request,
            _("The file could not be found. Please change the source or delete the audio file"),
            buttons=[messages.button(reverse("castaudio:delete", args=(audio.id,)), _("Delete"))],
        )

    return render(
        request,
        "cast/audio/edit.html",
        {
            "audio": audio,
            "filesize": filesize,
            "form": form,
            "user_can_delete": True,
        },
    )


def delete(request: HttpRequest, audio_id: int) -> HttpResponse:
    audio = get_object_or_404(Audio, id=audio_id)

    if request.POST:
        audio.delete()
        messages.success(request, _("Audio '{0}' deleted.").format(audio.title))
        return redirect("castaudio:index")

    return render(request, "cast/audio/confirm_delete.html", {"audio": audio})


def chooser(request: HttpRequest) -> HttpResponse:
    ordering = "-created"
    audios = Audio.objects.all().order_by(ordering)

    upload_form = AudioForm(prefix="media-chooser-upload")

    if "q" in request.GET or "p" in request.GET:
        search_form = NonEmptySearchForm(request.GET)
        if search_form.is_valid():
            q = search_form.cleaned_data["q"]

            audios = audios.search(q)
            is_searching = True
        else:
            q = None
            is_searching = False

        paginator, audios = paginate(request, audios, per_page=CHOOSER_PAGINATION)
        return render(
            request,
            "cast/audio/chooser_results.html",
            {
                "audios": audios,
                "query_string": q,
                "is_searching": is_searching,
                "pagination_template": pagination_template,
            },
        )
    else:
        search_form = NonEmptySearchForm()
        paginator, audios = paginate(request, audios, per_page=CHOOSER_PAGINATION)

    return render_modal_workflow(
        request,
        "cast/audio/chooser_chooser.html",
        None,
        {
            "audios": audios,
            "uploadform": upload_form,
            "searchform": search_form,
            "is_searching": False,
            "pagination_template": pagination_template,
        },
        json_data={
            "step": "chooser",
            "error_label": "Server Error",
            "error_message": "Report this error to your webmaster with the following information:",
            "tag_autocomplete_url": reverse("wagtailadmin_tag_autocomplete"),
        },
    )


def get_audio_data(audio: Audio) -> dict[str, Any]:
    """
    helper function: given a audio, return the json to pass back to the
    chooser panel - move to model FIXME
    """
    return {
        "id": audio.id,
        "title": audio.title,
        "edit_link": reverse("castaudio:edit", args=(audio.id,)),
    }


def chosen(request, audio_id: int) -> HttpResponse:
    audio = get_object_or_404(Audio, id=audio_id)

    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={"step": "audio_chosen", "result": get_audio_data(audio)},
    )


def chooser_upload(request: AuthenticatedHttpRequest) -> HttpResponse:
    if request.method == "POST":
        audio = Audio(user=request.user)
        form = AudioForm(request.POST, request.FILES, instance=audio, user=request.user, prefix="media-chooser-upload")

        if form.is_valid():
            form.save()

            # Reindex the media entry to make sure all tags are indexed
            for backend in get_search_backends():
                backend.add(audio)

            return render_modal_workflow(
                request,
                None,
                None,
                None,
                json_data={"step": "audio_chosen", "result": get_audio_data(audio)},
            )
        else:
            messages.error(request, _("The audio could not be saved due to errors."))

    ordering = "-created"
    audios = Audio.objects.all().order_by(ordering)

    search_form = NonEmptySearchForm()

    paginator, audios = paginate(request, audios, per_page=CHOOSER_PAGINATION)

    context = {
        "audios": audios,
        "searchform": search_form,
        # "collections": collections,
        "uploadform": AudioForm(),
        "is_searching": False,
        "pagination_template": "wagtailadmin/shared/ajax_pagination_nav.html",
    }
    return render_modal_workflow(
        request,
        "cast/audio/chooser_chooser.html",
        None,
        context,
        json_data={"step": "chooser"},
    )



# Generated by Django 3.2.6 on 2021-08-20 11:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0013_alter_gallery_images"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="gallery",
            name="user",
        ),
    ]



from unittest.mock import patch

import pytest
from django.urls import reverse

from cast.models import Audio
from cast.views.audio import delete_old_audio_files


class TestPostWithAudioDetail:
    pytestmark = pytest.mark.django_db

    def test_get_post_with_audio_detail(self, client, post_with_audio):
        audio = post_with_audio.audios.first()
        detail_url = post_with_audio.get_url()

        r = client.get(detail_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "html" in content

        # make sure audio is included in rendered audio block
        assert "block-audio" in content
        assert f"audio_{audio.pk}" in content


def get_endpoint_urls_without_args():
    urls = {}
    view_names = ["index", "add", "chooser", "chooser_upload"]
    for view_name in view_names:
        urls[view_name] = reverse(f"castaudio:{view_name}")
    return urls


def get_endpoint_urls_with_args(audio):
    urls = {}
    view_names = ["edit", "delete", "chosen"]
    for view_name in view_names:
        urls[view_name] = reverse(f"castaudio:{view_name}", args=(audio.id,))
    return urls


class AudioUrls:
    def __init__(self, audio):
        self.audio = audio
        self.urls = get_endpoint_urls_without_args()
        self.urls.update(get_endpoint_urls_with_args(audio))

    def __getattr__(self, item):
        return self.urls[item]


@pytest.fixture
def audio_urls(audio):
    return AudioUrls(audio)


class TestAllAudioEndpoints:
    pytestmark = pytest.mark.django_db

    def test_get_all_not_authenticated(self, client, audio_urls):
        for view_name, url in audio_urls.urls.items():
            r = client.get(url)

            # redirect to log in
            assert r.status_code == 302
            login_url = reverse("wagtailadmin_login")
            assert login_url in r.url

    def test_get_all_authenticated(self, authenticated_client, audio_urls):
        for view_name, url in audio_urls.urls.items():
            r = authenticated_client.get(url)

            # assert we are not redirected to log in
            assert r.status_code == 200


class TestAudioIndex:
    pytestmark = pytest.mark.django_db

    def test_get_index(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.index)

        assert r.status_code == 200
        content = r.content.decode("utf-8")

        # make sure it's the media results page
        assert "html" in content
        assert "media-results" in content

        # make sure audio_urls.audio is included in results
        assert audio_urls.audio.title in content

    def test_get_index_ajax(self, authenticated_client, audio_urls):
        headers = {"HTTP_X_REQUESTED_WITH": "XMLHttpRequest"}
        r = authenticated_client.get(audio_urls.index, **headers)

        assert r.status_code == 200
        content = r.content.decode("utf-8")

        # make sure it's the media results page
        assert "table" in content
        assert "listing" in content

        # make sure audio_urls.audio is included in results
        assert audio_urls.audio.title in content

    def test_get_index_with_search(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.index, {"q": audio_urls.audio.title})

        assert r.status_code == 200
        content = r.content.decode("utf-8")

        # make sure it's the media results page
        assert "html" in content
        assert "media-results" in content

        # make sure audio_urls.audio is included in results
        assert audio_urls.audio.title in content

    def test_get_index_with_search_invalid(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.index, {"q": " "})

        assert r.status_code == 200
        content = r.content.decode("utf-8")

        # make sure it's the media results page
        assert "html" in content
        assert "media-results" in content

        # make sure audio_urls.audio is included in results
        assert audio_urls.audio.title in content

    def test_get_index_with_pagination(self, authenticated_client, user):
        audio_models = []
        for i in range(1, 3):
            audio = Audio(user=user, title=f"audio {i}")
            audio.save()
            audio_models.append(audio)
        index_url = reverse("castaudio:index")
        with patch("cast.views.audio.MENU_ITEM_PAGINATION", return_value=1):
            r = authenticated_client.get(index_url, {"p": "1"})
        audios = r.context["audios"]

        # make sure we got last audio from first page
        assert len(audios) == 1
        assert audios[0] == audio_models[-1]

        with patch("cast.views.audio.MENU_ITEM_PAGINATION", return_value=1):
            r = authenticated_client.get(index_url, {"p": "2"})
        audios = r.context["audios"]

        # make sure we got first audio from last page
        assert len(audios) == 1
        assert audios[0] == audio_models[0]


class TestAudioAdd:
    pytestmark = pytest.mark.django_db

    def test_get_add_audio(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.add)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "Uploading…" in content

    def test_post_add_audio_invalid_form(self, authenticated_client, m4a_audio):
        m4a_audio.seek(m4a_audio.size)  # seek to end to make file empty/invalid
        add_url = reverse("castaudio:add")

        post_data = {
            "title": "foobar",
            "m4a": m4a_audio,  # invalid
            "tags": "foo,bar,baz",
        }
        r = authenticated_client.post(add_url, post_data)

        # make sure we don't get redirected to index
        assert r.status_code == 200
        assert r.context["message"] == "The audio file could not be saved due to errors."

        # make sure we didn't create an audio
        assert Audio.objects.first() is None

    def test_post_add_audio(self, authenticated_client, minimal_mp4):
        add_url = reverse("castaudio:add")

        post_data = {
            "title": "foobar",
            "tags": "foo,bar,baz",
            "original": minimal_mp4,
        }
        r = authenticated_client.post(add_url, post_data)

        # make sure we get redirected to index
        assert r.status_code == 302
        assert r.url == reverse("castaudio:index")

        # make sure field were saved correctly
        audio = Audio.objects.first()
        assert audio.title == post_data["title"]

        actual_tags = {t.name for t in audio.tags.all()}
        expected_tags = set(post_data["tags"].split(","))
        assert actual_tags == expected_tags


class TestAudioEdit:
    pytestmark = pytest.mark.django_db

    def test_get_edit_audio(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.edit)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "Delete" in content

    def test_get_edit_audio_without_m4a(self, authenticated_client, audio_without_m4a):
        audio = audio_without_m4a
        edit_url = reverse("castaudio:edit", args=(audio.id,))
        r = authenticated_client.get(edit_url)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "Delete" in content

    def test_get_edit_audio_without_m4a_no_filesize(self, settings, authenticated_client, audio_without_m4a):
        settings.DEFAULT_FILE_STORAGE = "django.core.files.storage.FileSystemStorage"
        audio = audio_without_m4a

        # set the name to make bool(audio.m4a) True and save it (yes, this is needed)
        audio.m4a.name = "foobar"
        audio.save()

        edit_url = reverse("castaudio:edit", args=(audio.id,))
        r = authenticated_client.get(edit_url)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "Delete" in content

    def test_post_edit_audio_invalid_form(self, authenticated_client, audio_urls, m4a_audio):
        m4a_audio.seek(m4a_audio.size)  # seek to end to make file empty/invalid
        post_data = {"m4a": m4a_audio}
        r = authenticated_client.post(audio_urls.edit, post_data)

        # make sure we don't get redirected to index
        assert r.status_code == 200

    def test_post_edit_audio_title(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        post_data = {
            "title": "changed title",
        }
        r = authenticated_client.post(audio_urls.edit, post_data)

        # make sure we get redirected to index
        assert r.status_code == 302
        assert r.url == audio_urls.index

        # make sure title was changes
        audio.refresh_from_db()
        assert audio.title == post_data["title"]

    def test_post_edit_audio_m4a(self, authenticated_client, audio_urls, m4a_audio):
        m4a_audio.seek(0)  # don't know why this is necessary :/
        post_data = {"m4a": m4a_audio}
        r = authenticated_client.post(audio_urls.edit, post_data)

        # make sure we get redirected to index
        assert r.status_code == 302
        assert r.url == audio_urls.index

        # teardown
        audio = Audio.objects.first()
        audio.m4a.delete()

    def test_delete_old_audio_files_skip_if_empty(self):
        class File:
            name = ""

        class OldAudio:
            mp3 = File()

        audio = OldAudio()
        assert delete_old_audio_files(audio, {"mp3"}) is None


class TestAudioDelete:
    pytestmark = pytest.mark.django_db

    def test_get_delete_audio(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.delete)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "Are you sure you want to delete this audio?" in content

    def test_post_delete_audio(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        # post data is necessary because of if request.POST
        r = authenticated_client.post(audio_urls.delete, {"delete": "yes"})

        # make sure we get redirected to index
        assert r.status_code == 302
        assert r.url == audio_urls.index

        # make sure audio was deleted
        with pytest.raises(Audio.DoesNotExist):
            audio.refresh_from_db()


class TestAudioChosen:
    pytestmark = pytest.mark.django_db

    def test_get_chosen_audio_not_found(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        audio.delete()
        r = authenticated_client.get(audio_urls.chosen)

        assert r.status_code == 404

    def test_get_chosen_audio_success(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        r = authenticated_client.get(audio_urls.chosen)

        assert r.status_code == 200

        # make sure returned data belongs to the right audio instance
        data = r.json()
        assert data["result"]["title"] == audio.title


class TestAudioChooser:
    pytestmark = pytest.mark.django_db

    def test_get_audio_in_chooser(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        r = authenticated_client.get(audio_urls.chooser)

        assert r.status_code == 200

        # make sure existing audio is in chooser
        content = r.content.decode("utf-8")
        assert audio.title in content

        # make sure prefix for form fields is set
        assert "media-chooser-upload" in content

    def test_get_chooser_with_search(self, authenticated_client, audio_urls):
        r = authenticated_client.get(audio_urls.chooser, {"q": audio_urls.audio.title})

        assert r.status_code == 200

        # make sure searched audio is included in results
        assert r.context["audios"][0] == audio_urls.audio

    def test_get_chooser_with_search_invalid(self, authenticated_client, audio_urls):
        # {"p": "1"} (page 1) leads to the search form being invalid
        r = authenticated_client.get(audio_urls.chooser, {"p": "1"})

        assert r.status_code == 200

        # make sure searched audios is included in results
        assert r.context["audios"][0] == audio_urls.audio

    def test_get_chooser_with_pagination(self, authenticated_client, user):
        audio_models = []
        for i in range(1, 3):
            audio = Audio(user=user, title=f"audio {i}")
            audio.save()
            audio_models.append(audio)
        chooser_url = reverse("castaudio:chooser")
        with patch("cast.views.audio.CHOOSER_PAGINATION", return_value=1):
            r = authenticated_client.get(chooser_url, {"p": "1"})
        audios = r.context["audios"]

        # make sure we got last audio from first page
        assert len(audios) == 1
        assert audios[0] == audio_models[-1]

        with patch("cast.views.audio.CHOOSER_PAGINATION", return_value=1):
            r = authenticated_client.get(chooser_url, {"p": "2"})
        audios = r.context["audios"]

        # make sure we got first audio from last page
        assert len(audios) == 1
        assert audios[0] == audio_models[0]


class TestAudioChooserUpload:
    pytestmark = pytest.mark.django_db

    def test_get_audio_in_chooser_upload(self, authenticated_client, audio_urls):
        audio = audio_urls.audio
        r = authenticated_client.get(audio_urls.chooser_upload)

        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert audio.title in content

    def test_post_upload_audio_form_invalid(self, authenticated_client, m4a_audio):
        m4a_audio.seek(m4a_audio.size)  # seek to end to make file empty/invalid
        upload_url = reverse("castaudio:chooser_upload")
        post_data = {"media-chooser-upload-m4a": m4a_audio}
        r = authenticated_client.post(upload_url, post_data)

        assert r.status_code == 200
        assert r.context["message"] == "The audio could not be saved due to errors."

    def test_post_upload_audio(self, authenticated_client, m4a_audio, settings):
        settings.DEFAULT_FILE_STORAGE = "django.core.files.storage.FileSystemStorage"
        upload_url = reverse("castaudio:chooser_upload")
        prefix = "media-chooser-upload"
        post_data = {
            f"{prefix}-title": "foobar",
            f"{prefix}-tags": "foo,bar,baz",
            f"{prefix}-m4a": m4a_audio,
        }
        r = authenticated_client.post(upload_url, post_data)

        assert r.status_code == 200

        # make sure field were saved correctly
        audio = Audio.objects.first()
        assert audio.title == post_data[f"{prefix}-title"]

        actual_tags = {t.name for t in audio.tags.all()}
        expected_tags = set(post_data[f"{prefix}-tags"].split(","))
        assert actual_tags == expected_tags

        # teardown
        audio.m4a.delete()



import io
import json
import os
import shutil
from copy import deepcopy
from datetime import datetime

import pytest
import pytz
from django.contrib.auth.models import Group
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test.client import RequestFactory
from django.utils import timezone
from django_comments import get_model as get_comments_model
from django_htmx.middleware import HtmxDetails
from rest_framework.test import APIClient
from wagtail.images.models import Image
from wagtail.models import Site

from cast import appsettings
from cast.models import Audio, ChapterMark, File, ItunesArtWork, Video

from .factories import (
    BlogFactory,
    EpisodeFactory,
    GalleryFactory,
    PodcastFactory,
    PostFactory,
    UserFactory,
    VideoFactory,
)


@pytest.fixture(scope="module")
def api_client():
    return APIClient()


@pytest.fixture()
def fixture_dir():
    current_directory = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(current_directory, "fixtures")


@pytest.fixture(scope="session", autouse=True)
def remove_stale_media_files():
    # runs before test starts
    yield
    # runs after test ends
    # cannot use function scoped settings fixture, so import settings
    from django.conf import settings

    shutil.rmtree(settings.MEDIA_ROOT, ignore_errors=True)


# Image testing stuff
def create_1_px_image():
    # This is a 1x1 black png
    png = (
        b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00"
        b"\x00\x01\x00\x00\x00\x01\x08\x06\x00\x00\x00"
        b"\x1f\x15\xc4\x89\x00\x00\x00\rIDATx\x9cc````"
        b"\x00\x00\x00\x05\x00\x01\xa5\xf6E@\x00\x00"
        b"\x00\x00IEND\xaeB`\x82"
    )
    return png


@pytest.fixture()
def image_1px():
    png = create_1_px_image()
    simple_png = SimpleUploadedFile(name="test.png", content=png, content_type="image/png")
    return simple_png


@pytest.fixture()
def image_1px_io():
    png = create_1_px_image()
    bio_file = io.BytesIO(png)
    bio_file.name = "test_image.png"
    bio_file.seek(0)
    return bio_file


def read_test_mp4(fixture_dir):
    with open(os.path.join(fixture_dir, "test.mp4"), "rb") as f:
        mp4 = f.read()
    return mp4


@pytest.fixture()
def minimal_mp4(fixture_dir):
    mp4 = read_test_mp4(fixture_dir)
    simple_mp4 = SimpleUploadedFile(name="test.mp4", content=mp4, content_type="video/mp4")
    return simple_mp4


def create_small_rgb():
    # this is a small test jpeg
    from PIL import Image

    img = Image.new("RGB", (200, 200), (255, 0, 0, 0))
    return img


@pytest.fixture()
def small_jpeg_io():
    rgb = create_small_rgb()
    im_io = io.BytesIO()
    rgb.save(im_io, format="JPEG", quality=60, optimize=True, progressive=True)
    im_io.name = "test_image.jpg"
    im_io.seek(0)
    return im_io


# Audio testing stuff
def create_minimal_mp3():
    mp3 = (
        b"\xff\xe3\x18\xc4\x00\x00\x00\x03H\x00\x00\x00\x00"
        b"LAME3.98.2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )
    return mp3


def read_test_m4a(fixture_dir):
    with open(os.path.join(fixture_dir, "test.m4a"), "rb") as f:
        m4a = f.read()
    return m4a


@pytest.fixture()
def mp3_audio():
    mp3 = create_minimal_mp3()
    simple_mp3 = SimpleUploadedFile(name="test.mp3", content=mp3, content_type="audio/mpeg")
    return simple_mp3


@pytest.fixture()
def m4a_audio(fixture_dir):
    m4a = read_test_m4a(fixture_dir)
    simple_m4a = SimpleUploadedFile(name="test.m4a", content=m4a, content_type="audio/mp4")
    return simple_m4a


# Models
@pytest.fixture()
def user(settings):
    user = UserFactory()
    user._password = "password"
    group = Group.objects.get(name="Moderators")
    group.user_set.add(user)
    return user


@pytest.fixture()
def authenticated_client(client, user):
    client.login(username=user.username, password=user._password)
    return client


@pytest.fixture()
def image(image_1px):
    image = Image(file=image_1px)
    image.save()
    return image


@pytest.fixture()
def video_with_poster(user, minimal_mp4, image_1px):
    video = Video(user=user, original=minimal_mp4, poster=image_1px)
    video.save()
    return video


@pytest.fixture()
def itunes_artwork(image_1px):
    ia = ItunesArtWork(original=image_1px)
    ia.save()
    return ia


@pytest.fixture()
def audio(user, m4a_audio, settings):
    settings.DEFAULT_FILE_STORAGE = "django.core.files.storage.FileSystemStorage"
    audio = Audio(user=user, m4a=m4a_audio, title="foobar audio")
    audio.save()
    teardown_path = audio.m4a.path  # save path to unlink if audio.m4a is set to None
    yield audio
    # teardown
    try:
        os.unlink(teardown_path)
    except FileNotFoundError:
        pass


@pytest.fixture()
def chaptermarks(audio):
    cms = [
        ("00:01:01.234", "introduction", "", ""),
        ("00:03:05.567", "coughing", "https://google.com", ""),
        ("00:02:05.567", "wrong order", "", ""),
    ]
    results = []
    for start, title, href, image in cms:
        results.append(ChapterMark.objects.create(audio=audio, start=start, title=title))
    return results


@pytest.fixture()
def file_instance(user, m4a_audio):
    file_instance = File(user=user, original=m4a_audio)
    file_instance.save()
    return file_instance


@pytest.fixture()
def site():
    return Site.objects.first()


@pytest.fixture()
def blog(user, site):
    return BlogFactory(owner=user, title="test blog", slug="test_blog", parent=site.root_page)


@pytest.fixture()
def podcast(user, site):
    return PodcastFactory(owner=user, title="test podcast", slug="test_podcast", parent=site.root_page)


@pytest.fixture()
def podcast_with_artwork(user, itunes_artwork, site):
    return PodcastFactory(
        owner=user,
        title="test podcast",
        slug="test_podcast",
        itunes_artwork=itunes_artwork,
        parent=site.root_page,
    )


@pytest.fixture()
def podcast_with_itunes_categories(user, site):
    categories = {"foo": ["baz"]}
    return PodcastFactory(
        owner=user,
        title="test blog",
        slug="test_blog",
        itunes_categories=json.dumps(categories),
        parent=site.root_page,
    )


@pytest.fixture()
def post_data():
    return {"title": "foobar", "content": "blub", "explicit": "2"}


@pytest.fixture()
def post_data_wagtail():
    return {
        "action-publish": "action-publish",
        "body-0-deleted": "",
        "body-0-order": "0",
        "body-0-type": "overview",
        "body-0-value-0-deleted": "",
        "body-0-value-0-order": "0",
        "body-0-value-0-type": "heading",
        "body-0-value-0-value": "overview heading",
        "body-0-value-count": "1",
        "body-count": "1",
        "slug": "new-post",
        "title": "new post",
        "visible_date": "2021-08-17 08:13",
    }


@pytest.fixture()
def python_body():
    return [
        {
            "type": "overview",
            "value": [
                {
                    "type": "heading",
                    "value": "in_all heading",
                }
            ],
        },
        {
            "type": "detail",
            "value": [
                {
                    "type": "heading",
                    "value": "only_in_detail heading",
                }
            ],
        },
    ]


@pytest.fixture()
def body(python_body):
    return json.dumps(python_body)


@pytest.fixture()
def body_with_gallery(python_body, gallery):
    image_pks = [img.pk for img in gallery.images.all()]
    gallery_body = deepcopy(python_body)
    gallery_with_layout = {"layout": "default", "gallery": image_pks}
    gallery_body[0]["value"].append({"type": "gallery", "value": gallery_with_layout})
    return json.dumps(gallery_body)


@pytest.fixture
def body_with_video(python_body, video):
    video_body = deepcopy(python_body)
    video_body[0]["value"].append({"type": "video", "value": video.id})
    return json.dumps(video_body)


@pytest.fixture
def body_with_audio(python_body, audio):
    audio_body = deepcopy(python_body)
    audio_body[0]["value"].append({"type": "audio", "value": audio.id})
    return json.dumps(audio_body)


@pytest.fixture
def body_with_image(python_body, image):
    image_body = deepcopy(python_body)
    image_body[0]["value"].append({"type": "image", "value": image.id})
    return json.dumps(image_body)


@pytest.fixture()
def post(blog, body):
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        body=body,
    )


@pytest.fixture()
def post_in_podcast(podcast, body):
    return PostFactory(
        owner=podcast.owner,
        parent=podcast,
        title="test entry",
        slug="test-entry",
        body=body,
    )


@pytest.fixture()
def post_with_gallery(blog, body_with_gallery):
    print("body with gallery: ", body_with_gallery)
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        body=body_with_gallery,
    )


@pytest.fixture
def post_with_video(blog, body_with_video):
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        body=body_with_video,
    )


@pytest.fixture
def post_with_audio(blog, body_with_audio):
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        body=body_with_audio,
    )


@pytest.fixture
def post_with_image(blog, body_with_image):
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        body=body_with_image,
    )


@pytest.fixture()
def unpublished_post(blog):
    post = PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
    )
    post.unpublish()
    post.refresh_from_db()
    return post


@pytest.fixture()
def unpublished_episode_without_audio(episode):
    episode.podcast_audio = None
    episode.unpublish()
    episode.refresh_from_db()
    return episode


@pytest.fixture()
def post_with_date(blog):
    visible_date = pytz.timezone("Europe/Berlin").localize(datetime(2018, 1, 1, 8))
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry",
        visible_date=visible_date,
    )


@pytest.fixture()
def post_with_different_date(blog):
    visible_date = pytz.timezone("Europe/Berlin").localize(datetime(2019, 1, 1, 8))
    return PostFactory(
        owner=blog.owner,
        parent=blog,
        title="test entry",
        slug="test-entry-alt",
        visible_date=visible_date,
    )


@pytest.fixture()
def post_with_search(blog, python_body):
    search_body = deepcopy(python_body)
    query = "only_in_search"
    search_body[-1]["value"][0]["value"] = f"{query} foobar"
    body = json.dumps(search_body)
    post = PostFactory(
        owner=blog.owner,
        parent=blog,
        title="asdf",
        slug="test-entry-with-search",
        visible_date=timezone.now(),
        body=body,
    )
    post.query = query
    return post


@pytest.fixture()
def episode(podcast, audio, body):
    return EpisodeFactory(
        owner=podcast.owner,
        parent=podcast,
        title="test podcast episode",
        slug="test-podcast-entry",
        podcast_audio=audio,
        body=body,
    )


@pytest.fixture()
def podcast_episode_with_same_audio(blog, audio, body):
    return EpisodeFactory(
        owner=blog.owner,
        parent=blog,
        title="test podcast episode 2",
        slug="test-podcast-entry2",
        podcast_audio=audio,
        body=body,
    )


@pytest.fixture()
def podcast_episode_with_different_visible_date(podcast, audio):
    visible_date = pytz.timezone("Europe/Berlin").localize(datetime(2019, 1, 1, 8))
    return EpisodeFactory(
        owner=podcast.owner,
        parent=podcast,
        title="test podcast episode",
        slug="test-podcast-entry",
        visible_date=visible_date,
        podcast_audio=audio,
    )


@pytest.fixture()
def test_templ():
    return """
        {% lorem %}
    """


@pytest.fixture()
def img_templ():
    return """
        {{% load cast_extras %}}
        {{% image {} %}}
    """


@pytest.fixture()
def video(user):
    video = VideoFactory.build()
    video.title = "video title"
    video.user = user
    video.save(poster=False)
    return video


@pytest.fixture
def video_without_original(video):
    video.original = None
    video.save()
    return video


@pytest.fixture
def audio_without_m4a(audio):
    audio.m4a = None
    audio.save()
    return audio


@pytest.fixture
def video_without_file(video):
    video.original.delete()
    video.original.name = "does_not_exist"
    video.save()
    return video


@pytest.fixture()
def gallery(image):
    gallery = GalleryFactory()
    gallery.images.add(image)
    return gallery


# Mocks
class DummyHandler:
    def __init__(self):
        self.se = {}
        self.ee = set()
        self.aqe = {}

    def startElement(self, name, value):
        self.se[name] = value

    def endElement(self, name):
        self.ee.add(name)

    def addQuickElement(self, *args, attrs=None):
        name = args[0]
        value = None
        if len(args) > 1:
            value = args[1]
        self.aqe[name] = (value, attrs)


@pytest.fixture()
def dummy_handler():
    return DummyHandler()


@pytest.fixture()
def request_factory():
    return RequestFactory()


@pytest.fixture
def simple_request(request_factory):
    request = request_factory.get("/")
    request.htmx = HtmxDetails(request)
    request.session = {}
    return request


@pytest.fixture
def htmx_request(request_factory):
    request = request_factory.get("/", HTTP_HX_REQUEST="true", HTTP_HX_TARGET="paging-area")
    request.htmx = HtmxDetails(request)
    return request


@pytest.fixture
def htmx_request_without_target(request_factory):
    request = request_factory.get("/", HTTP_HX_REQUEST="true", HTTP_HX_TARGET=None)
    request.htmx = HtmxDetails(request)
    return request


@pytest.fixture()
def comments_enabled():
    previous = appsettings.CAST_COMMENTS_ENABLED
    appsettings.CAST_COMMENTS_ENABLED = True
    yield appsettings.CAST_COMMENTS_ENABLED
    appsettings.CAST_COMMENTS_ENABLED = previous


@pytest.fixture()
def comments_not_enabled():
    previous = appsettings.CAST_COMMENTS_ENABLED
    appsettings.CAST_COMMENTS_ENABLED = False
    yield appsettings.CAST_COMMENTS_ENABLED
    appsettings.CAST_COMMENTS_ENABLED = previous


@pytest.fixture()
def comment(post, settings):
    comment_model = get_comments_model()
    instance = comment_model(content_object=post, site_id=settings.SITE_ID, title="foobar", comment="bar baz")
    instance.save()
    return instance


@pytest.fixture()
def comment_spam(post, settings):
    comment_model = get_comments_model()
    instance = comment_model(
        content_object=post, site_id=settings.SITE_ID, title="blub", comment="asdf bsdf", is_removed=True
    )
    instance.save()
    return instance


@pytest.fixture
def use_dummy_cache_backend(settings, mocker):
    settings.CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    }
    # workaround for settings fixture is not working in Django 4.0 and pytest-django
    mocker.patch("django.core.cache.backends.locmem.LocMemCache.get", return_value=None)



from typing import Union

from django.conf import settings
from django.db.models.signals import post_delete
from wagtail.images import get_image_model
from wagtail.images.signal_handlers import post_delete_file_cleanup

CAST_COMMENTS_ENABLED: bool = getattr(settings, "CAST_COMMENTS_ENABLED", False)
CAST_CUSTOM_THEMES: list[tuple[str, str]] = []
CHOOSER_PAGINATION: int = getattr(settings, "CHOOSER_PAGINATION", 10)
MENU_ITEM_PAGINATION: int = getattr(settings, "MENU_ITEM_PAGINATION", 20)
POST_LIST_PAGINATION: int = getattr(settings, "POST_LIST_PAGINATION", 5)
DELETE_WAGTAIL_IMAGES: bool = getattr(settings, "DELETE_WAGTAIL_IMAGES", True)
CAST_FILTERSET_FACETS: list[str] = getattr(
    settings, "CAST_FILTERSET_FACETS", ["search", "date", "date_facets", "category_facets", "tag_facets", "o"]
)
CAST_IMAGE_FORMATS: list[str] = getattr(settings, "CAST_IMAGE_FORMATS", ["jpeg", "avif"])
CAST_REGULAR_IMAGE_SLOT_DIMENSIONS: list[tuple[int, int]] = getattr(
    settings, "CAST_REGULAR_IMAGE_SLOT_DIMENSIONS", [(1110, 740)]
)
CAST_GALLERY_IMAGE_SLOT_DIMENSIONS: list[tuple[int, int]] = getattr(
    settings, "CAST_GALLERY_IMAGE_SLOT_DIMENSIONS", [(1110, 740), (120, 80)]
)

SettingValue = Union[str, bool, int]


def set_default_if_not_set(setting_name: str, default_value: SettingValue) -> None:
    if getattr(settings, setting_name, None) is None:
        setattr(settings, setting_name, default_value)


DEFAULT_VALUES: list[tuple[str, SettingValue]] = [
    ("SITE_ID", 1),
    ("WAGTAIL_SITE_NAME", "Cast"),
    ("CRISPY_TEMPLATE_PACK", "bootstrap4"),
    ("CRISPY_ALLOWED_TEMPLATE_PACKS", "bootstrap4"),
]


def init_cast_settings() -> None:
    if not DELETE_WAGTAIL_IMAGES:
        # Have a way to deactivate wagtails post_delete_file_cleanup
        # which deletes the file physically when developing against S3
        # cast has to be after wagtail in INSTALLED_APPS for this to work
        post_delete.disconnect(post_delete_file_cleanup, sender=get_image_model())

    for setting_name, default_value in DEFAULT_VALUES:
        set_default_if_not_set(setting_name, default_value)



# Generated by Django 3.2.6 on 2021-08-20 09:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0023_add_choose_permissions"),
        ("cast", "0012_alter_post_images"),
    ]

    operations = [
        migrations.AlterField(
            model_name="gallery",
            name="images",
            field=models.ManyToManyField(to="wagtailimages.Image"),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-24 06:13

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0052_pagelogentry"),
        ("cast", "0004_homepage_alias_for_page"),
    ]

    operations = [
        migrations.AlterField(
            model_name="homepage",
            name="alias_for_page",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Make this page an alias for another page, redirecting to it with a non permanent redirect.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="aliases",
                to="wagtailcore.page",
                verbose_name="Redirect to another page",
            ),
        ),
    ]



import json
from typing import Optional, Union

from django import forms
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.widgets import BaseChooser, BaseChooserAdapter
from wagtail.telepath import register

from .models import Audio, Video


def add_display_none_to_chooser_button(html: str) -> str:
    """
    Workaround for duplicate chooser button in wagtail admin.
    """
    return mark_safe(html.replace("data-chooser-action-choose", "data-chooser-action-choose style='display: none'"))


class CastChooser(BaseChooser):
    chooser_namespace = "castvideo"
    template_name = "cast/video/chooser.html"

    @property
    def chooser_modal_url_name(self) -> str:
        return f"{self.chooser_namespace}:chooser"

    def get_chooser_modal_url(self) -> str:
        return reverse(self.chooser_modal_url_name)

    def render_html(self, name: str, value: Optional[dict], attrs: dict) -> str:
        value = value if value is not None else {}
        original_field_html = super().render_html(name, value.get("id"), attrs)
        original_field_html = add_display_none_to_chooser_button(original_field_html)

        context = {
            "widget": self,
            "original_field_html": original_field_html,
            "attrs": attrs,
            "value": value != {},  # only used to identify blank values
            "title": value.get("title", ""),
            "edit_url": value.get("edit_url", ""),
        }
        html = render_to_string(self.template_name, context=context)
        return html


class AdminVideoChooser(CastChooser):
    choose_one_text = _("Choose a video item")
    choose_another_text = _("Choose another video item")
    link_to_chosen_text = _("Edit this video item")

    def get_value_data(self, value: Optional[Union[Video, int]]) -> Optional[dict]:
        if value is None:
            return value
        if not isinstance(value, Video):
            value = Video.objects.get(pk=value)
        assert isinstance(value, Video)
        return {
            "id": value.pk,
            "title": value.title,
            "edit_link": reverse("castvideo:edit", args=[value.id]),
        }

    def render_html(self, name: str, value: Optional[dict], attrs: dict) -> str:
        value = value if value is not None else {}
        original_field_html = super().render_html(name, value.get("id"), attrs)
        original_field_html = add_display_none_to_chooser_button(original_field_html)

        context = {
            "widget": self,
            "original_field_html": original_field_html,
            "attrs": attrs,
            "value": value != {},  # only used to identify blank values
            "title": value.get("title", ""),
            "edit_url": value.get("edit_url", ""),
        }
        html = render_to_string("cast/video/chooser.html", context=context)
        return html

    def render_js_init(self, id_: int, name: str, value: Optional[dict]) -> str:
        return f"createVideoChooser({json.dumps(id_)});"

    class Media:
        js = [
            "cast/js/wagtail/video-chooser-modal.js",
            "cast/js/wagtail/video-chooser.js",
        ]


class VideoChooserAdapter(BaseChooserAdapter):
    js_constructor = "cast.wagtail.VideoChooser"

    def js_args(self, widget: AdminVideoChooser) -> list:
        return [
            widget.render_html("__NAME__", None, attrs={"id": "__ID__"}),
            widget.id_for_label("__ID__"),
        ]

    @cached_property
    def media(self) -> forms.Media:
        return forms.Media(
            js=[
                versioned_static("cast/js/wagtail/video-chooser-telepath.js"),
            ]
        )


register(VideoChooserAdapter(), AdminVideoChooser)


class AdminAudioChooser(CastChooser):
    choose_one_text = _("Choose an audio item")
    choose_another_text = _("Choose another audio item")
    link_to_chosen_text = _("Edit this audio item")
    chooser_namespace = "castaudio"
    template_name = "cast/audio/chooser.html"

    def get_value_data(self, value: Optional[Union["Video", int]]) -> Optional[dict]:
        if value is None:
            return value
        if not isinstance(value, Audio):
            value = Audio.objects.get(pk=value)
        assert isinstance(value, Audio)
        return {
            "id": value.pk,
            "title": value.title,
            "edit_link": reverse("castaudio:edit", args=[value.id]),
        }

    def render_js_init(self, id_, name: str, value: Optional[dict]) -> str:
        return f"createAudioChooser({json.dumps(id_)});"

    class Media:
        js = [
            "cast/js/wagtail/audio-chooser-modal.js",
            "cast/js/wagtail/audio-chooser.js",
        ]


class AudioChooserAdapter(BaseChooserAdapter):
    js_constructor = "cast.wagtail.AudioChooser"

    def js_args(self, widget: AdminAudioChooser) -> list:
        return [
            widget.render_html("__NAME__", None, attrs={"id": "__ID__"}),
            widget.id_for_label("__ID__"),
        ]

    @cached_property
    def media(self) -> forms.Media:
        return forms.Media(
            js=[
                versioned_static("cast/js/wagtail/audio-chooser-telepath.js"),
            ]
        )


register(AudioChooserAdapter(), AdminAudioChooser)



from django.core.files.storage import FileSystemStorage
from django.core.management.base import BaseCommand

from .storage_backend import get_production_and_backup_storage


class Command(BaseCommand):
    help = (
        "replace paths on production storage backend with local versions - useful for compressed videos for example"
        "(requires Django >= 4.2 and production and backup storage configured)"
    )

    def add_arguments(self, parser):
        parser.add_argument("paths", nargs="+", type=str)

    def handle(self, *args, **options):
        production, _ = get_production_and_backup_storage()
        fs_storage = FileSystemStorage()
        for path in options["paths"]:
            if fs_storage.exists(path):
                if production.exists(path):
                    production.delete(path)
                with fs_storage.open(path, "rb") as in_f:
                    production.save(path, in_f)



from collections.abc import Iterable
from typing import Optional

from django.db import models
from model_utils.models import TimeStampedModel
from wagtail.images.models import Image


class Gallery(TimeStampedModel):
    images: models.ManyToManyField = models.ManyToManyField(Image)  # FIXME mypy are you ok?
    post_context_key = "gallery"

    @property
    def image_ids(self) -> set[int]:
        return {i.pk for i in self.images.all()}


def get_or_create_gallery(image_ids: Iterable[int]) -> Optional[Gallery]:
    candidate_images = Image.objects.filter(id__in=image_ids)  # FIXME filter permissions
    if candidate_images.count() == 0:
        return None
    filtered_image_ids = [ci.id for ci in candidate_images]
    gallery_to_image_ids = {}
    # FIXME filter permissions - fetch only images / galleries that
    # this user has permission to view
    candidate_galleries = Gallery.objects.filter(images__in=filtered_image_ids).prefetch_related("images")
    for gallery in candidate_galleries:
        gallery_to_image_ids[frozenset(i.id for i in gallery.images.all())] = gallery
    gallery = gallery_to_image_ids.get(frozenset(filtered_image_ids))
    if gallery is None:
        gallery = Gallery.objects.create()
        gallery.images.add(*filtered_image_ids)
    return gallery



# Generated by Django 4.1.7 on 2023-03-12 21:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0042_blog_template_base_dir"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blog",
            name="template_base_dir",
            field=models.CharField(
                blank=True,
                choices=[("bootstrap4", "Bootstrap 4"), ("plain", "Just HTML")],
                default=None,
                max_length=10,
                null=True,
            ),
        ),
    ]



# Generated by Django 4.0a1 on 2021-10-05 11:53

import cast.models.moderation
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0022_alter_spamfilter_model"),
    ]

    operations = [
        migrations.AlterField(
            model_name="spamfilter",
            name="model",
            field=models.JSONField(
                decoder=cast.models.moderation.ModelDecoder,
                default=dict,
                encoder=cast.models.moderation.ModelEncoder,
                verbose_name="Spamfilter Model",
            ),
        ),
    ]



import logging
from datetime import datetime

from django.contrib.syndication.views import Feed
from django.db.models import QuerySet
from django.http import Http404, HttpRequest
from django.shortcuts import get_object_or_404
from django.utils.feedgenerator import Atom1Feed, Rss201rev2Feed, rfc2822_date
from django.utils.safestring import SafeText

from .models import Audio, Blog, Episode, Podcast, Post

logger = logging.getLogger(__name__)


class LatestEntriesFeed(Feed):
    object: Blog
    request: HttpRequest

    def get_object(self, request: HttpRequest, *args, **kwargs) -> None:
        slug = kwargs["slug"]
        self.object = get_object_or_404(Blog, slug=slug)
        self.request = request

    def title(self) -> str:
        return self.object.title

    def description(self) -> str:
        return self.object.description

    def link(self) -> str:
        return self.object.get_full_url()

    def items(self) -> QuerySet[Post]:
        queryset = Post.objects.live().descendant_of(self.object).order_by("-visible_date")
        return queryset

    def item_title(self, item) -> SafeText:
        return item.title

    def item_description(self, item) -> SafeText:
        item.description = item.get_description(request=self.request, render_detail=True, escape_html=False)
        return item.description

    def item_link(self, item) -> SafeText:
        return item.get_full_url()


class ITunesElements:
    feed: dict

    def add_artwork(self, podcast: Podcast, handler) -> None:
        if podcast.itunes_artwork is None:
            return

        haqe = handler.addQuickElement
        itunes_artwork_url = podcast.itunes_artwork.original.url
        handler.addQuickElement("itunes:image", attrs={"href": itunes_artwork_url})
        handler.startElement("image", {})
        haqe("url", itunes_artwork_url)
        haqe("title", self.feed["title"])
        handler.endElement("image")

    def add_itunes_categories(self, podcast: Podcast, handler) -> None:
        itunes_categories = podcast.itunes_categories_parsed
        if len(itunes_categories) == 0:
            return
        for category, subcategories in itunes_categories.items():
            handler.startElement("itunes:category", {"text": category})
            for subcategory in subcategories:
                handler.addQuickElement("itunes:category", attrs={"text": subcategory})
            handler.endElement("itunes:category")

    def add_root_elements(self, handler):
        """Add additional elements to the blog object"""
        super().add_root_elements(handler)
        haqe = handler.addQuickElement
        blog = self.feed["blog"]

        self.add_artwork(blog, handler)

        haqe("itunes:subtitle", self.feed["subtitle"])
        haqe("itunes:author", blog.author_name)
        handler.startElement("itunes:owner", {})
        haqe("itunes:name", blog.author_name)
        haqe("itunes:email", blog.email)
        handler.endElement("itunes:owner")

        self.add_itunes_categories(blog, handler)

        haqe("itunes:summary", blog.description)
        haqe("itunes:explicit", blog.get_explicit_display())
        try:
            haqe("lastBuildDate", rfc2822_date(blog.last_build_date))
        except IndexError:
            pass
        generator = "Django Web Framework / django-cast"
        haqe("generator", generator)
        haqe("docs", "https://blogs.law.harvard.edu/tech/rss")

    def add_item_elements(self, handler, item):
        """Add additional elements to the post object"""
        super().add_item_elements(handler, item)
        haqe = handler.addQuickElement

        post = item["post"]
        haqe("guid", str(post.uuid), attrs={"isPermaLink": "false"})
        # Maybe add license later
        # year = timezone.now().year
        # haqe("copyright", "{0} {1}".format("insert license", year))
        haqe("itunes:author", post.owner.get_full_name())
        haqe("itunes:subtitle", post.title)
        haqe("itunes:summary", post.description)
        haqe("itunes:duration", post.podcast_audio.duration_str)
        haqe("itunes:keywords", post.keywords)
        haqe("itunes:explicit", post.get_explicit_display())
        if post.block:
            haqe("itunes:block", "yes")

    def namespace_attributes(self):
        return {"xmlns:itunes": "http://www.itunes.com/dtds/podcast-1.0.dtd"}


class AtomITunesFeedGenerator(ITunesElements, Atom1Feed):
    def root_attributes(self):
        atom_attrs = super().root_attributes()
        atom_attrs.update(self.namespace_attributes())
        return atom_attrs


class RssITunesFeedGenerator(ITunesElements, Rss201rev2Feed):
    def rss_attributes(self):
        rss_attrs = super().rss_attributes()
        rss_attrs.update(self.namespace_attributes())
        return rss_attrs


class PodcastFeed(Feed):
    """
    A feed of podcasts for iTunes and other compatible podcatchers.
    """

    audio_format: str
    mime_type: str
    object: Podcast
    request: HttpRequest

    def set_audio_format(self, audio_format: str) -> None:
        format_to_mime = Audio.mime_lookup
        if audio_format not in format_to_mime:
            raise Http404("unknown audio format")
        else:
            self.audio_format = audio_format
            self.mime_type = format_to_mime[audio_format]

    def get_object(self, request, *args, **kwargs):
        self.set_audio_format(kwargs["audio_format"])

        slug = kwargs["slug"]
        self.object = get_object_or_404(Podcast, slug=slug)
        self.request = request  # need request for item.serve(request) later on
        return self.object

    def link(self) -> str:
        return self.object.get_full_url()

    def title(self, blog: Blog) -> str:
        return self.object.title

    def categories(self, blog: Blog) -> tuple[str]:
        if hasattr(blog, "categories"):
            return (blog.keywords.split(",")[0],)
        else:
            return ("",)

    def itunes_categories(self, blog: Blog) -> list[str]:
        return blog.itunes_categories.split(",")

    def items(self, podcast: Podcast) -> QuerySet[Episode]:
        queryset = (
            Episode.objects.live()
            .descendant_of(podcast)
            .filter(podcast_audio__isnull=False)
            .select_related("podcast_audio")
            .order_by("-visible_date")
        )
        return queryset

    def item_title(self, item):
        return item.title

    def item_description(self, item):
        item.description = item.get_description(request=self.request, render_detail=True, escape_html=False)
        return item.description

    def item_link(self, item):
        return item.get_full_url()

    def item_pubdate(self, item) -> datetime:
        return item.visible_date

    def item_updateddate(self, item: Post) -> datetime:
        return item.last_published_at

    # def item_categories(self, post):
    #    return self.categories(self.blog)

    def item_enclosure_url(self, item: Post) -> str:
        return item.get_enclosure_url(self.audio_format)

    def item_enclosure_length(self, item: Post) -> int:
        return item.get_enclosure_size(self.audio_format)

    def item_enclosure_mime_type(self, _item: Post) -> str:
        return self.mime_type

    def item_keywords(self, item: Post) -> str:
        return item.keywords

    def feed_extra_kwargs(self, obj) -> dict:
        return {"blog": self.object}

    def item_extra_kwargs(self, item):
        return {"blog": self.object, "post": item}


class AtomPodcastFeed(PodcastFeed):
    feed_type = AtomITunesFeedGenerator

    def subtitle(self, blog: Blog) -> str:
        return blog.description

    def author_name(self, blog: Blog) -> str:
        return blog.author_name

    def author_email(self, blog):
        return blog.email

    def link(self) -> str:
        """atom link is still wrong, dunno why FIXME"""
        return self.object.get_full_url()


class RssPodcastFeed(PodcastFeed):
    feed_type = RssITunesFeedGenerator

    def item_guid(self, _post: Post) -> None:
        """ITunesElements can't add isPermaLink attr unless None is returned here."""
        return None

    def description(self, blog: Blog) -> str:
        return blog.description



from django.http import HttpRequest
from django_htmx.middleware import HtmxDetails


# Typing pattern recommended by django-stubs:
# https://github.com/typeddjango/django-stubs#how-can-i-create-a-httprequest-thats-guaranteed-to-have-an-authenticated-user
class HtmxHttpRequest(HttpRequest):
    htmx: HtmxDetails



from typing import Any

from django.urls import include, path
from django.views.decorators.cache import cache_page

from . import feeds
from .views import meta
from .views.gallery import gallery_modal
from .views.theme import select_theme

app_name = "cast"
urlpatterns: list[Any] = [
    # API
    path("api/", include("cast.api.urls", namespace="api")),
    # Feeds
    path(
        "<slug:slug>/feed/rss.xml",
        view=cache_page(5 * 60)(feeds.LatestEntriesFeed()),
        name="latest_entries_feed",
    ),
    path(
        "<slug:slug>/feed/podcast/<audio_format>/rss.xml",
        view=cache_page(5 * 60)(feeds.RssPodcastFeed()),
        name="podcast_feed_rss",
    ),
    path(
        "<slug:slug>/feed/podcast/<audio_format>/atom.xml",
        view=cache_page(5 * 60)(feeds.AtomPodcastFeed()),
        name="podcast_feed_atom",
    ),
    # Meta views like twitter player cards etc
    path("<slug:blog_slug>/<slug:episode_slug>/twitter-player/", view=meta.twitter_player, name="twitter-player"),
    # Store selected theme in session
    path("select-theme/", view=select_theme, name="select-theme"),
    # Gallery modal via htmx
    path("gallery_modal/<str:template_base_dir>/", view=gallery_modal, name="gallery-modal"),
]



import json
import logging
import re
import subprocess
from collections.abc import Iterable
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Protocol, runtime_checkable

from django.contrib.auth import get_user_model
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel
from taggit.managers import TaggableManager
from wagtail.models import CollectionMember, PageManager
from wagtail.search import index
from wagtail.search.queryset import SearchableQuerySetMixin

if TYPE_CHECKING:
    from .pages import Episode

logger = logging.getLogger(__name__)


class AudioQuerySet(SearchableQuerySetMixin, models.QuerySet):
    pass


@runtime_checkable
class FileField(Protocol):
    """Just to make mypy happy about field.url and field.path"""

    url: str
    path: str


class Audio(CollectionMember, index.Indexed, TimeStampedModel):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    duration = models.DurationField(null=True, blank=True)
    title = models.CharField(max_length=255, null=True, blank=True)
    subtitle = models.CharField(max_length=512, null=True, blank=True)

    m4a = models.FileField(upload_to="cast_audio/", null=True, blank=True)
    mp3 = models.FileField(upload_to="cast_audio/", null=True, blank=True)
    oga = models.FileField(upload_to="cast_audio/", null=True, blank=True)
    opus = models.FileField(upload_to="cast_audio/", null=True, blank=True)

    data = models.JSONField("Metadata", blank=True, default=dict)

    mime_lookup: dict[str, str] = {
        "mp3": "audio/mpeg",
        "m4a": "audio/mp4",
        "oga": "audio/ogg",
        "opus": "audio/opus",
    }
    audio_formats: list[str] = list(mime_lookup.keys())
    title_lookup: dict[str, str] = {key: f"Audio {key.upper()}" for key in audio_formats}

    admin_form_fields: tuple[str, ...] = ("title", "subtitle", "m4a", "mp3", "oga", "opus", "tags")

    search_fields = CollectionMember.search_fields + [
        index.SearchField("title", partial_match=True, boost=10),
        index.RelatedFields(
            "tags",
            [
                index.SearchField("title", partial_match=True, boost=10),
                index.SearchField("subtitle", partial_match=True, boost=5),
            ],
        ),
        index.FilterField("user"),
    ]

    objects: PageManager["Audio"] = AudioQuerySet.as_manager()
    tags = TaggableManager(help_text=None, blank=True, verbose_name=_("tags"))

    @property
    def uploaded_audio_files(self) -> Iterable[tuple[str, models.FileField]]:
        for name in self.audio_formats:
            field = getattr(self, name)
            if field.name is not None and len(field.name) > 0:
                yield name, field

    @property
    def file_formats(self) -> str:
        return " ".join([n for n, f in self.uploaded_audio_files])

    def get_audio_file_names(self) -> set[str]:
        audio_file_names = set()
        for audio_format, field in self.uploaded_audio_files:
            audio_file_names.add(Path(field.name).stem)
        return audio_file_names

    @property
    def name(self) -> str:
        if self.title is not None:
            return self.title
        return ",".join([_ for _ in self.get_audio_file_names()])

    def __str__(self) -> str:
        return f"{self.pk} - {self.name}"

    def get_all_paths(self) -> set[str]:
        paths = set()
        for name, field in self.uploaded_audio_files:
            paths.add(field.name)
        return paths

    @staticmethod
    def _get_audio_duration(audio_url) -> timedelta:
        # Taken from: http://trac.ffmpeg.org/wiki/FFprobeTips
        cmd = f"""
        ffprobe  \
            -v 0  \
            -print_format json  \
            -show_entries format=duration  \
            -of default=noprint_wrappers=1:nokey=1  \
            '{audio_url}'
        """
        result = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT).decode().strip()
        m = re.match(r"^(?P<seconds>\d+)\.(?P<microseconds>\d+)$", result)
        if m is None:
            raise ValueError(f"Could not parse duration: {result}")
        return timedelta(seconds=int(m["seconds"]), microseconds=int(m["microseconds"]))

    def create_duration(self) -> None:
        for name, field in self.uploaded_audio_files:
            try:
                # For mypy this has to be a direct isinstance check
                # In production it raises a NotImplementedError, very weird,
                # very ugly, dunno how to fix it
                if isinstance(field, FileField):
                    audio_url = field.url
                    if not audio_url.startswith("http"):
                        audio_url = field.path
                    self.duration = self._get_audio_duration(audio_url)
                    break
            except NotImplementedError:  # pragma: no cover
                pass

    @property
    def audio(self) -> list[dict[str, str]]:
        items = []
        for name, field in self.uploaded_audio_files:
            if not hasattr(field, "url"):
                continue
            items.append(
                {
                    "url": field.url,  # type: ignore
                    "mimeType": self.mime_lookup[name],
                    "size": str(self.get_file_size(name)),
                    "title": str(self.title_lookup[name]),
                }
            )
        return items

    @property
    def chapters(self) -> list[dict[str, str]]:
        items = []
        # chapter marks have to be ordered by start for
        # podlove web player - dunno why, 2019-04-19 jochen
        for chapter in self.chaptermarks.order_by("start"):
            items.append(
                {
                    "start": str(chapter.start).split(".")[0],
                    "title": chapter.title,
                    "href": chapter.link,
                    "image": chapter.image,
                }
            )
        return items

    @property
    def chapters_as_text(self) -> str:
        chaptermarks = []
        for mark in self.chaptermarks.order_by("start"):
            chaptermarks.append(mark.original_line)
        return "\n".join(chaptermarks)

    @staticmethod
    def clean_ffprobe_chaptermarks(ffprobe_data: dict) -> list[dict[str, str]]:
        cleaned = []
        for item in ffprobe_data["chapters"]:
            start = item["start_time"]
            title = item["tags"]["title"]
            if title == "":
                continue
            cleaned.append({"start": start, "title": title})
        return cleaned

    def get_chaptermark_data_from_file(self, audio_format: str) -> list[dict[str, str]]:
        file_field = getattr(self, audio_format)
        try:
            url = file_field.url
        except ValueError:
            return []
        if not url.startswith("http"):
            # use path from local filesystem
            url = file_field.path
        command = [
            "ffprobe",
            "-i",
            str(url),
            "-print_format",
            "json",
            "-show_chapters",
            "-loglevel",
            "error",
        ]
        ffprobe_data = json.loads(subprocess.run(command, check=True, stdout=subprocess.PIPE).stdout)
        return self.clean_ffprobe_chaptermarks(ffprobe_data)

    def set_episode_id(self, episode_id: int) -> None:
        """Set the episode id for this audio file to be able to return audio.episode_url in api."""
        self._episode_id = episode_id

    def get_episode(self, episode_id=None) -> Optional["Episode"]:
        episodes = self.episodes.all()
        if episode_id is not None:
            episodes = episodes.filter(pk=episode_id)
        episodes = list(episodes)
        if len(episodes) == 1:
            return episodes[0]
        return None

    @property
    def episode_url(self) -> Optional[str]:
        """Return the url to the episode this audio file belongs to."""
        episode_id = None
        if hasattr(self, "_episode_id"):
            episode_id = self._episode_id
        episode = self.get_episode(episode_id)
        if episode is not None:
            return episode.full_url
        return None

    @property
    def podlove_url(self) -> str:
        return reverse("cast:api:audio_podlove_detail", kwargs={"pk": self.pk})

    @property
    def duration_str(self) -> str:
        dur = str(self.duration)
        return dur.split(".")[0]

    def size_to_metadata(self) -> None:
        self.data["size"] = self.data.get("size", {})
        for audio_format, field in self.uploaded_audio_files:
            try:
                assert hasattr(field, "size"), f"field {field} has no size attribute"
                self.data["size"][audio_format] = field.size
            except FileNotFoundError:
                # file does not exist -> do not cache
                pass

    def get_file_size(self, audio_format: str) -> int:
        """Return the file size of the given audio format."""
        cached_size = self.data.get("size", {}).get(audio_format)
        if cached_size is not None:
            return cached_size
        file_field = getattr(self, audio_format)
        try:
            assert hasattr(file_field, "size"), f"field {file_field} has no size attribute"
            return file_field.size
        except ValueError:
            # file_field is null
            return 0

    def save(self, *args, **kwargs) -> None:
        generate_duration = kwargs.pop("duration", True)
        cache_file_sizes = kwargs.pop("cache_file_sizes", True)
        # FIXME why is this necessary? Cannot move super save to end of method...
        super().save(*args, **kwargs)
        if generate_duration and self.duration is None:
            logger.info("save audio duration")
            self.create_duration()
            super().save(*args, **kwargs)
        if cache_file_sizes:
            self.size_to_metadata()
            super().save(*args, **kwargs)


def sync_chapter_marks(
    from_database: list["ChapterMark"], from_cms: list["ChapterMark"]
) -> tuple[list["ChapterMark"], list["ChapterMark"], list["ChapterMark"]]:
    start_from_database = {cm.start: cm for cm in from_database}
    start_from_cms = {cm.start for cm in from_cms}
    to_add, to_update = [], []
    for cm in from_cms:
        if cm.start not in start_from_database:
            to_add.append(cm)
        else:
            cm_from_db = start_from_database[cm.start]
            if cm.has_changed(cm_from_db):
                cm.pk = cm_from_db.pk  # to be able to just call cm.save() later on
                to_update.append(cm)
    to_remove = [cm for cm in from_database if cm.start not in start_from_cms]
    return to_add, to_update, to_remove


class ChapterMarkManager(models.Manager):
    @staticmethod
    def sync_chaptermarks(audio: Audio, from_cms: list["ChapterMark"]) -> None:
        from_db: list["ChapterMark"] = list(audio.chaptermarks.all())
        to_add, to_update, to_remove = sync_chapter_marks(from_db, from_cms)
        for cm in to_add + to_update:
            cm.save()
        for cm in to_remove:
            cm.delete()


class ChapterMark(models.Model):
    audio = models.ForeignKey(Audio, on_delete=models.CASCADE, related_name="chaptermarks")
    start = models.TimeField(_("Start time of chaptermark"))
    title = models.CharField(max_length=255)
    link = models.URLField(max_length=2000, null=True, blank=True)
    image = models.URLField(max_length=2000, null=True, blank=True)

    objects = ChapterMarkManager()

    class Meta:
        unique_together = (("audio", "start"),)

    def __str__(self) -> str:
        return f"{self.pk} {self.start} {self.title}"

    def has_changed(self, other: "ChapterMark") -> bool:
        return self.title != other.title

    @property
    def original_line(self) -> str:
        link = ""
        if self.link is not None:
            link = self.link
        image = ""
        if self.image is not None:
            image = self.image
        return f"{self.start} {self.title} {link} {image}"
